from PySide6.QtCore import Qt, QAbstractItemModel, QRect
from PySide6.QtWidgets import QTableView, QHeaderView
from PySide6.QtGui import QResizeEvent, QPainter

# ----------------------------------------------------------------------------------------------------------------------
# File implements TableViewWithFooter class that is a descendant of QTableView class with footer.
# Underlying model should support footerData(section, role) method in order to provide data for the footer.
# The footer is implemented as FooterView class that is derived from QHeaderView.
# ----------------------------------------------------------------------------------------------------------------------
class FooterView(QHeaderView):
    def __init__(self, parent: QTableView):
        super().__init__(Qt.Horizontal, parent)
        self._model = None
        self.setSectionResizeMode(QHeaderView.Fixed)

    def setModel(self, model: QAbstractItemModel) -> None:
        self._model = model
        self._model.dataChanged.connect(self.on_model_update)
        super().setModel(model)

    def paintSection(self, painter: QPainter, rect: QRect, idx: int) -> None:
        # First make standard painting by parent QHeaderView class method
        painter.save()
        super().paintSection(painter, rect, idx)
        painter.restore()
        # Clean footer content (by default QHeaderView puts sections names there
        inner_rect = rect.adjusted(self.lineWidth(), self.lineWidth(), -self.lineWidth(), -self.lineWidth())
        bg_color = self.palette().color(self.backgroundRole())
        painter.fillRect(inner_rect, bg_color)
        # Get data from model and write text with given font and position
        text = self._model.footerData(idx, role=Qt.DisplayRole)
        if text is None:
            return
        font = self._model.footerData(idx, role=Qt.FontRole)
        if font is not None:
            painter.setFont(font)
        alignment = self._model.footerData(idx, role=Qt.TextAlignmentRole)
        alignment = Qt.AlignCenter | Qt.AlignVCenter if alignment is None else alignment
        painter.drawText(inner_rect, alignment, f" {text} ")

    def on_header_resize(self, section: int, _old_size: int, new_size: int) -> None:
        self.resizeSection(section, new_size)

    def on_header_move(self, _section: int, old: int, new: int) -> None:
        self.moveSection(old, new)

    def on_model_update(self, top_left, bottom_right, _role) -> None:
        self.headerDataChanged(Qt.Horizontal, top_left.column(), bottom_right.column())


class TableViewWithFooter(QTableView):
    def __init__(self, parent_view):
        self._parent_view = parent_view
        super().__init__(parent_view)
        self._footer = FooterView(self)
        self._header = self.horizontalHeader()
        self._header.sectionResized.connect(self._footer.on_header_resize)
        self._header.sectionMoved.connect(self._footer.on_header_move)

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        m = self.viewportMargins()
        self.setViewportMargins(m.left(), m.top(), m.right(), m.top())  # Mirror top margin to bottom
        cr = self.contentsRect()
        header_size = self._header.geometry()
        self._footer.setGeometry(cr.left(), cr.top() + cr.height() - m.top() + 1, header_size.width(), header_size.height())

    def setModel(self, model: QAbstractItemModel) -> None:
        super().setModel(model)
        self._footer.setModel(model)

    def setColumnHidden(self, idx: int, hidden: bool) -> None:
        super().setColumnHidden(idx, hidden)
        self._footer.setSectionHidden(idx, hidden)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dividend_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDateEdit, QDateTimeEdit,
    QGridLayout, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpacerItem, QWidget)

from jal.widgets.account_select import AccountCurrencyLabel
from jal.widgets.reference_selector import (AccountSelector, AssetSelector)

class Ui_DividendOperation(object):
    def setupUi(self, DividendOperation):
        if not DividendOperation.objectName():
            DividendOperation.setObjectName(u"DividendOperation")
        DividendOperation.resize(968, 195)
        self.layout = QGridLayout(DividendOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.note = QLineEdit(DividendOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 4, 1, 1, 9)

        self.date_label = QLabel(DividendOperation)
        self.date_label.setObjectName(u"date_label")
        self.date_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.date_label, 1, 0, 1, 1)

        self.number_label = QLabel(DividendOperation)
        self.number_label.setObjectName(u"number_label")
        self.number_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.number_label, 1, 7, 1, 1)

        self.commit_button = QPushButton(DividendOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 11, 1, 1)

        self.price_edit = QLineEdit(DividendOperation)
        self.price_edit.setObjectName(u"price_edit")
        self.price_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.price_edit.setReadOnly(True)

        self.layout.addWidget(self.price_edit, 2, 8, 1, 1)

        self.type_label = QLabel(DividendOperation)
        self.type_label.setObjectName(u"type_label")
        self.type_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.type_label, 1, 5, 1, 1)

        self.tax_label = QLabel(DividendOperation)
        self.tax_label.setObjectName(u"tax_label")
        self.tax_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.tax_label, 3, 5, 1, 1)

        self.number = QLineEdit(DividendOperation)
        self.number.setObjectName(u"number")

        self.layout.addWidget(self.number, 1, 8, 1, 1)

        self.symbol_label = QLabel(DividendOperation)
        self.symbol_label.setObjectName(u"symbol_label")
        self.symbol_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.symbol_label, 3, 0, 1, 1)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.vertical_spacer, 5, 0, 1, 1)

        self.main_label = QLabel(DividendOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 1)

        self.currency = AccountCurrencyLabel(DividendOperation)
        self.currency.setObjectName(u"currency")

        self.layout.addWidget(self.currency, 3, 7, 1, 1)

        self.horizontal_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontal_spacer, 2, 10, 1, 1)

        self.tax_edit = QLineEdit(DividendOperation)
        self.tax_edit.setObjectName(u"tax_edit")
        self.tax_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.tax_edit, 3, 6, 1, 1)

        self.ex_date_editor = QDateEdit(DividendOperation)
        self.ex_date_editor.setObjectName(u"ex_date_editor")
        self.ex_date_editor.setMinimumDate(QDate(1970, 1, 1))
        self.ex_date_editor.setMaximumTime(QTime(23, 59, 59))
        self.ex_date_editor.setCalendarPopup(True)

        self.layout.addWidget(self.ex_date_editor, 1, 3, 1, 1)

        self.account_widget = AccountSelector(DividendOperation)
        self.account_widget.setObjectName(u"account_widget")

        self.layout.addWidget(self.account_widget, 2, 1, 1, 4)

        self.asset_widget = AssetSelector(DividendOperation)
        self.asset_widget.setObjectName(u"asset_widget")

        self.layout.addWidget(self.asset_widget, 3, 1, 1, 4)

        self.note_label = QLabel(DividendOperation)
        self.note_label.setObjectName(u"note_label")
        self.note_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.note_label, 4, 0, 1, 1)

        self.account_label = QLabel(DividendOperation)
        self.account_label.setObjectName(u"account_label")
        self.account_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.account_label, 2, 0, 1, 1)

        self.ex_date_label = QLabel(DividendOperation)
        self.ex_date_label.setObjectName(u"ex_date_label")

        self.layout.addWidget(self.ex_date_label, 1, 2, 1, 1)

        self.revert_button = QPushButton(DividendOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 12, 1, 1)

        self.amount_label = QLabel(DividendOperation)
        self.amount_label.setObjectName(u"amount_label")
        self.amount_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.amount_label, 2, 5, 1, 1)

        self.dividend_edit = QLineEdit(DividendOperation)
        self.dividend_edit.setObjectName(u"dividend_edit")
        self.dividend_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.dividend_edit, 2, 6, 1, 1)

        self.type = QComboBox(DividendOperation)
        self.type.setObjectName(u"type")

        self.layout.addWidget(self.type, 1, 6, 1, 1)

        self.timestamp_editor = QDateTimeEdit(DividendOperation)
        self.timestamp_editor.setObjectName(u"timestamp_editor")
        self.timestamp_editor.setCalendarPopup(True)
        self.timestamp_editor.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.timestamp_editor, 1, 1, 1, 1)

        self.price_label = QLabel(DividendOperation)
        self.price_label.setObjectName(u"price_label")

        self.layout.addWidget(self.price_label, 2, 7, 1, 1)


        self.retranslateUi(DividendOperation)

        QMetaObject.connectSlotsByName(DividendOperation)
    # setupUi

    def retranslateUi(self, DividendOperation):
        DividendOperation.setWindowTitle(QCoreApplication.translate("DividendOperation", u"Form", None))
        self.date_label.setText(QCoreApplication.translate("DividendOperation", u"Date/Time", None))
        self.number_label.setText(QCoreApplication.translate("DividendOperation", u"#", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("DividendOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
        self.type_label.setText(QCoreApplication.translate("DividendOperation", u"Type", None))
        self.tax_label.setText(QCoreApplication.translate("DividendOperation", u"Tax", None))
        self.symbol_label.setText(QCoreApplication.translate("DividendOperation", u"Asset", None))
        self.main_label.setText(QCoreApplication.translate("DividendOperation", u"Dividend", None))
        self.currency.setText(QCoreApplication.translate("DividendOperation", u"CUR", None))
        self.ex_date_editor.setSpecialValueText(QCoreApplication.translate("DividendOperation", u"unknown", None))
        self.ex_date_editor.setDisplayFormat(QCoreApplication.translate("DividendOperation", u"dd/MM/yyyy", None))
        self.note_label.setText(QCoreApplication.translate("DividendOperation", u"Note", None))
        self.account_label.setText(QCoreApplication.translate("DividendOperation", u"Account", None))
        self.ex_date_label.setText(QCoreApplication.translate("DividendOperation", u"Ex-Date", None))
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("DividendOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
        self.amount_label.setText(QCoreApplication.translate("DividendOperation", u"Dividend", None))
        self.timestamp_editor.setDisplayFormat(QCoreApplication.translate("DividendOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.price_label.setText(QCoreApplication.translate("DividendOperation", u"Price", None))
    # retranslateUi




import re
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QDialog
from jal.data_import.receipt_api.ru_fns import ReceiptRuFNS
from jal.data_import.receipt_api.eu_lidl_plus import ReceiptEuLidlPlus
from jal.data_import.receipt_api.pt_pingo_doce import ReceiptPtPingoDoce
from jal.widgets.qr_scanner import ScanDialog, QRScanner


# ----------------------------------------------------------------------------------------------------------------------
# Possible values that may be used by factory
RU_FNS_API = 'RU_FNS'
EU_LIDL_PLUS_API = 'EU_LIDL_PLUS'
PT_PINGO_DOCE_API = 'PT_PINGO_DOCE'


# ----------------------------------------------------------------------------------------------------------------------
class ReceiptAPIFactory(QObject):
    def __init__(self):
        super().__init__()
        self._apis = {
            RU_FNS_API: ReceiptRuFNS,
            EU_LIDL_PLUS_API: ReceiptEuLidlPlus,
            PT_PINGO_DOCE_API: ReceiptPtPingoDoce
        }
        self.supported_names = {
            RU_FNS_API: self.tr("Russian receipt"),
            EU_LIDL_PLUS_API: self.tr("European Lidl receipt"),
            PT_PINGO_DOCE_API: self.tr("Portuguese Pingo Doce receipt")
        }
        self._pt_nifs = {
            '503340855': EU_LIDL_PLUS_API,
            '500829993': PT_PINGO_DOCE_API
        }

    def get_api_parameters(self, api_type):
        api = self._apis.get(api_type)
        return api.parameters_list()

    # Selects required API class based on QR data pattern
    def get_api_for_qr(self, qr_text):
        extra_data=''
        api_type = self._detect_api_id_by_qr(qr_text)
        if api_type == EU_LIDL_PLUS_API:
            scanner = ScanDialog(code_type=QRScanner.TYPE_ITF,
                                 message=self.tr("Please scan flat barcode from the receipt"))
            if scanner.exec() == QDialog.Accepted:
                extra_data = scanner.data
        api = self._apis.get(api_type)
        return api(qr_text=qr_text, aux_data=extra_data)

    def _detect_api_id_by_qr(self, qr_text):
        ru_fns_keys = ["i=", "n=", "s=", "t=", "fn=", "fp="]
        pt_at_pattern = r"A:(?P<NIF>.{1,9})\*B:.{1,30}\*C:PT\*D:FS\*E:N\*F:\d{8}\*G:.{1,30}\*H:.{1,70}\*I1:PT\*.*"
        if all([x in qr_text for x in ru_fns_keys]):
            return RU_FNS_API
        parts = re.match(pt_at_pattern, qr_text)
        if parts is not None:
            NIF = parts.groupdict()['NIF']
            try:
                return self._pt_nifs[NIF]
            except KeyError:
                raise ValueError(self.tr("Portuguese QR recognized but shop isn't supported, NIF: ") + f"{NIF}")
        raise ValueError(self.tr("No API found for QR data: ") + f"'{qr_text}'")

    def get_api_with_params(self, api_type, params):
        api = self._apis.get(api_type)
        return api(params=params)



import logging
from PySide6.QtCore import Qt, Slot, Signal, QRectF, QTimer, QMetaObject
from PySide6.QtGui import QImage, QPen, QBrush
from PySide6.QtWidgets import QApplication, QDialog, QWidget, QFrame, QVBoxLayout, QHBoxLayout, QSpacerItem, \
    QGraphicsScene, QGraphicsView, QLabel, QSizePolicy, QPushButton, QFileDialog, QInputDialog
from jal.widgets.helpers import dependency_present, decodeQR
try:
    from pyzbar import pyzbar
except ImportError:
    pass  # pyzbar import will be checked separately
try:
    from PySide6.QtMultimedia import QMediaDevices, QCamera, QMediaCaptureSession, QImageCapture
    from PySide6.QtMultimediaWidgets import QGraphicsVideoItem
except ImportError:
    pass   # We should not be in this module as dependencies have been checked in main_window.py and calls are disabled

# ----------------------------------------------------------------------------------------------------------------------
class QRScanner(QWidget):
    TYPE_QR = 1
    TYPE_ITF = 2
    QR_SIZE = 0.75      # Size of rectangle for QR capture (used to display only currently) or ITF width
    ITF_SIZE = 0.2      # Size of ITF height area for capture
    QR_SCAN_RATE = 100  # Delay in ms between QR captures
    decodedQR = Signal(str)

    def __init__(self, parent=None, code_type=TYPE_QR):
        super().__init__(parent=parent)
        self.processing = False
        self.started = False
        self.rectangle = None
        self.code_type = pyzbar.ZBarSymbol.I25 if code_type == self.TYPE_ITF else pyzbar.ZBarSymbol.QRCODE

        self.setMinimumHeight(405)
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.scene = QGraphicsScene(self)
        self.scene.setBackgroundBrush(QBrush(Qt.black))
        self.view = QGraphicsView(self.scene)
        self.viewfinder = QGraphicsVideoItem()
        self.scene.addItem(self.viewfinder)
        self.layout.addWidget(self.view)
        self.setLayout(self.layout)

        self.camera = None
        self.captureSession = None
        self.imageCapture = None
        self.captureTimer = None

    def startScan(self):
        if self.started:
            return
        if len(QMediaDevices.videoInputs()) == 0:
            logging.warning(self.tr("There are no cameras available"))
            return
        if not dependency_present(['pyzbar']):
            logging.warning(self.tr("Package pyzbar not found for QR recognition."))
            return ''

        self.processing = True   # disable any capture while camera is starting
        self.camera = QCamera(QMediaDevices.defaultVideoInput(), parent=self)
        self.captureSession = QMediaCaptureSession(self)
        self.imageCapture = QImageCapture(self)
        self.captureSession.setCamera(self.camera)
        self.captureSession.setVideoOutput(self.viewfinder)
        self.captureSession.setImageCapture(self.imageCapture)
        self.captureTimer = QTimer(self)

        self.camera.errorOccurred.connect(self.onCameraError)
        self.imageCapture.errorOccurred.connect(self.onCaptureError)
        self.imageCapture.imageCaptured.connect(self.onImageCaptured)
        self.viewfinder.nativeSizeChanged.connect(self.onVideoSizeChanged)
        self.captureTimer.timeout.connect(self.scanQR)

        self.camera.start()
        self.processing = False
        self.started = True
        self.captureTimer.start(self.QR_SCAN_RATE)

    def stopScan(self):
        if self.camera is None:
            return
        if not self.started:
            return
        self.processing = True  # disable capture
        self.captureTimer.stop()
        self.camera.stop()

        self.captureTimer.deleteLater()
        self.captureSession.deleteLater()
        self.imageCapture.deleteLater()
        self.camera.deleteLater()
        self.started = False

    def onVideoSizeChanged(self, _size):
        self.resizeEvent(None)

    # Take QImage or QRect (object with 'width' and 'height' properties and calculate position and size
    # of the square with side of self.QR_SIZE from minimum of height or width for QR code type
    # or the rectangle of QR_SIZE x ITF_SIZE for ITF code type
    def calculate_scan_area(self, img_rect) -> QRectF:
        min_side = min(img_rect.height(), img_rect.width())
        h = self.QR_SIZE * min_side    # Square side or code width
        w = self.ITF_SIZE * min_side if self.code_type == pyzbar.ZBarSymbol.I25 else h  # Code height
        x = (img_rect.width() - w) / 2         # Position of the square inside rectangle
        y = (img_rect.height() - h) / 2
        if type(img_rect) != QImage:   # if we have a bounding rectangle, not an image
            x += img_rect.left()       # then we need to shift our square inside this rectangle
            y += img_rect.top()
        return QRectF(x, y, w, h)

    def resizeEvent(self, event):
        bounds = self.scene.itemsBoundingRect()
        if bounds.width() <= 0 or bounds.height() <= 0:
            return    # do nothing if size is zero
        self.view.fitInView(bounds, Qt.KeepAspectRatio)
        if self.rectangle is not None:
            self.scene.removeItem(self.rectangle)
        pen = QPen(Qt.red) if self.code_type == pyzbar.ZBarSymbol.I25 else QPen(Qt.green)
        pen.setWidth(1)
        pen.setStyle(Qt.DashLine)
        self.rectangle = self.scene.addRect(self.calculate_scan_area(bounds), pen)
        self.view.centerOn(0, 0)
        self.view.raise_()

    def onCaptureError(self, _id, error, error_str):
        self.processing = False
        self.onCameraError(error, error_str)

    def onCameraError(self, error, error_str):
        logging.error(self.tr("Camera error: " + str(error) + " / " + error_str))

    def scanQR(self):
        if self.imageCapture is None:
            return
        if self.processing:
            return
        if self.imageCapture.isReadyForCapture():
            self.imageCapture.capture()
            self.processing = True

    def onImageCaptured(self, _id: int, img: QImage):
        self.decodeQR(img)
        self.processing = False

    def decodeQR(self, qr_image: QImage):
        cropped = qr_image.copy(self.calculate_scan_area(qr_image).toRect())
        qr_text = decodeQR(cropped, code_type=self.code_type)
        if qr_text:
            self.decodedQR.emit(qr_text)


# ----------------------------------------------------------------------------------------------------------------------
class ScanDialog(QDialog):
    def __init__(self, parent=None, code_type=QRScanner.TYPE_QR, message=''):
        super().__init__(parent)
        self.data = ''
        self.code_type = code_type
        self.running = False
        self.resize(700, 450)
        self.setWindowTitle(self.tr("Barcode scanner"))
        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.setSpacing(2)
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.hint_label = QLabel(message, parent=self)
        self.hint_label.setAlignment(Qt.AlignCenter)
        self.verticalLayout.addWidget(self.hint_label)

        self.scanner = QRScanner(self, self.code_type)
        self.scanner.decodedQR.connect(self.barcode_scanned)
        scanner_size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        scanner_size_policy.setHeightForWidth(self.scanner.sizePolicy().hasHeightForWidth())
        self.scanner.setSizePolicy(scanner_size_policy)
        self.verticalLayout.addWidget(self.scanner)
        self.ButtonFrame = QFrame(self)
        self.horizontalLayout = QHBoxLayout(self.ButtonFrame)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.readQR_button = QPushButton(self.ButtonFrame, text=self.tr("Load image from file"))
        self.pasteQR_button = QPushButton(self.ButtonFrame, text=self.tr("Get image from clipboard"))
        self.enterText_button = QPushButton(self.ButtonFrame, text=self.tr("Input data manually"))
        self.close_button = QPushButton(self.ButtonFrame, text=self.tr("Close"))
        self.h_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.horizontalLayout.addWidget(self.readQR_button)
        self.horizontalLayout.addWidget(self.pasteQR_button)
        self.horizontalLayout.addWidget(self.enterText_button)
        self.horizontalLayout.addItem(self.h_spacer)
        self.horizontalLayout.addWidget(self.close_button)
        self.verticalLayout.addWidget(self.ButtonFrame)
        self.readQR_button.clicked.connect(self.load_qr_image_file)
        self.pasteQR_button.clicked.connect(self.paste_qr_image)
        self.enterText_button.clicked.connect(self.enter_qr_data)
        self.close_button.clicked.connect(self.close)

    @Slot()
    def showEvent(self, event):
        super().showEvent(event)
        if self.running:
            return
        self.running = True
        # Call slot via queued connection, so it's called from the UI thread after the window has been shown
        QMetaObject().invokeMethod(self, "start_scan", Qt.ConnectionType.QueuedConnection)

    @Slot()
    def closeEvent(self, event):
        self.scanner.stopScan()

    @Slot()
    def start_scan(self):
        self.scanner.startScan()

    @Slot()
    def barcode_scanned(self, decoded_data):
        self.scanner.stopScan()
        self.data = decoded_data
        self.accept()

    #------------------------------------------------------------------------------------------
    # Loads graphics file and tries to read QR-code from it.
    @Slot()
    def load_qr_image_file(self):
        qr_file, _filter = \
            QFileDialog.getOpenFileName(self, self.tr("Select file with QR code"),
                                        ".", "JPEG images (*.jpg);;PNG images (*.png);;BMP images (*.bmp)")
        if qr_file:
            self.data = decodeQR(QImage(qr_file), code_type=self.code_type)
            if self.data:
                self.accept()
            else:
                logging.warning(self.tr("No QR codes were found in file"))

    # ------------------------------------------------------------------------------------------
    # Gets image from clipboard and tries to read QR-code from it.
    @Slot()
    def paste_qr_image(self):
        self.data = decodeQR(QApplication.clipboard().image(), code_type=self.code_type)
        if self.data:
            self.accept()
        else:
            logging.warning(self.tr("No QR codes found in clipboard"))

    @Slot()
    def enter_qr_data(self):
        self.data, result = QInputDialog.getText(None, self.tr("Input code data manually"), self.tr("Data:"))
        if not result:
            return
        self.accept()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trade_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDateEdit, QDateTimeEdit, QGridLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QWidget)

from jal.widgets.account_select import AccountCurrencyLabel
from jal.widgets.reference_selector import (AccountSelector, AssetSelector)

class Ui_TradeOperation(object):
    def setupUi(self, TradeOperation):
        if not TradeOperation.objectName():
            TradeOperation.setObjectName(u"TradeOperation")
        TradeOperation.resize(968, 210)
        self.layout = QGridLayout(TradeOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.price_label = QLabel(TradeOperation)
        self.price_label.setObjectName(u"price_label")
        self.price_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.price_label, 2, 7, 1, 1)

        self.qty_edit = QLineEdit(TradeOperation)
        self.qty_edit.setObjectName(u"qty_edit")
        self.qty_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.qty_edit, 1, 8, 1, 1)

        self.commit_button = QPushButton(TradeOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 11, 1, 1)

        self.price_edit = QLineEdit(TradeOperation)
        self.price_edit.setObjectName(u"price_edit")
        self.price_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.price_edit, 2, 8, 1, 1)

        self.currency_fee = AccountCurrencyLabel(TradeOperation)
        self.currency_fee.setObjectName(u"currency_fee")

        self.layout.addWidget(self.currency_fee, 3, 9, 1, 1)

        self.date_label = QLabel(TradeOperation)
        self.date_label.setObjectName(u"date_label")

        self.layout.addWidget(self.date_label, 1, 0, 1, 1)

        self.symbol_label = QLabel(TradeOperation)
        self.symbol_label.setObjectName(u"symbol_label")

        self.layout.addWidget(self.symbol_label, 3, 0, 1, 1)

        self.main_label = QLabel(TradeOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 1)

        self.settlement_label = QLabel(TradeOperation)
        self.settlement_label.setObjectName(u"settlement_label")

        self.layout.addWidget(self.settlement_label, 1, 2, 1, 1)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontalSpacer, 2, 10, 1, 1)

        self.comment_label = QLabel(TradeOperation)
        self.comment_label.setObjectName(u"comment_label")

        self.layout.addWidget(self.comment_label, 4, 0, 1, 1)

        self.account_label = QLabel(TradeOperation)
        self.account_label.setObjectName(u"account_label")

        self.layout.addWidget(self.account_label, 2, 0, 1, 1)

        self.fee_label = QLabel(TradeOperation)
        self.fee_label.setObjectName(u"fee_label")
        self.fee_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.fee_label, 3, 7, 1, 1)

        self.qty_label = QLabel(TradeOperation)
        self.qty_label.setObjectName(u"qty_label")
        self.qty_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.qty_label, 1, 7, 1, 1)

        self.currency_price = AccountCurrencyLabel(TradeOperation)
        self.currency_price.setObjectName(u"currency_price")

        self.layout.addWidget(self.currency_price, 2, 9, 1, 1)

        self.fee_edit = QLineEdit(TradeOperation)
        self.fee_edit.setObjectName(u"fee_edit")
        self.fee_edit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.fee_edit, 3, 8, 1, 1)

        self.revert_button = QPushButton(TradeOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 12, 1, 1)

        self.number_label = QLabel(TradeOperation)
        self.number_label.setObjectName(u"number_label")
        self.number_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.number_label, 1, 5, 1, 1)

        self.timestamp_editor = QDateTimeEdit(TradeOperation)
        self.timestamp_editor.setObjectName(u"timestamp_editor")
        self.timestamp_editor.setCalendarPopup(True)
        self.timestamp_editor.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.timestamp_editor, 1, 1, 1, 1)

        self.settlement_editor = QDateEdit(TradeOperation)
        self.settlement_editor.setObjectName(u"settlement_editor")
        self.settlement_editor.setCalendarPopup(True)

        self.layout.addWidget(self.settlement_editor, 1, 3, 1, 1)

        self.account_widget = AccountSelector(TradeOperation)
        self.account_widget.setObjectName(u"account_widget")

        self.layout.addWidget(self.account_widget, 2, 1, 1, 6)

        self.note = QLineEdit(TradeOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 4, 1, 1, 9)

        self.asset_widget = AssetSelector(TradeOperation)
        self.asset_widget.setObjectName(u"asset_widget")

        self.layout.addWidget(self.asset_widget, 3, 1, 1, 6)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.vertical_spacer, 5, 0, 1, 1)

        self.number = QLineEdit(TradeOperation)
        self.number.setObjectName(u"number")

        self.layout.addWidget(self.number, 1, 6, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontalSpacer_2, 1, 4, 1, 1)


        self.retranslateUi(TradeOperation)

        QMetaObject.connectSlotsByName(TradeOperation)
    # setupUi

    def retranslateUi(self, TradeOperation):
        TradeOperation.setWindowTitle(QCoreApplication.translate("TradeOperation", u"Form", None))
        self.price_label.setText(QCoreApplication.translate("TradeOperation", u"Price", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("TradeOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
        self.currency_fee.setText(QCoreApplication.translate("TradeOperation", u"CUR", None))
        self.date_label.setText(QCoreApplication.translate("TradeOperation", u"Date/Time", None))
        self.symbol_label.setText(QCoreApplication.translate("TradeOperation", u"Asset", None))
        self.main_label.setText(QCoreApplication.translate("TradeOperation", u"Buy / Sell", None))
        self.settlement_label.setText(QCoreApplication.translate("TradeOperation", u"Settlement", None))
        self.comment_label.setText(QCoreApplication.translate("TradeOperation", u"Note", None))
        self.account_label.setText(QCoreApplication.translate("TradeOperation", u"Account", None))
        self.fee_label.setText(QCoreApplication.translate("TradeOperation", u"Fee", None))
        self.qty_label.setText(QCoreApplication.translate("TradeOperation", u"Qty", None))
        self.currency_price.setText(QCoreApplication.translate("TradeOperation", u"CUR", None))
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("TradeOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
        self.number_label.setText(QCoreApplication.translate("TradeOperation", u"#", None))
        self.timestamp_editor.setDisplayFormat(QCoreApplication.translate("TradeOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.settlement_editor.setDisplayFormat(QCoreApplication.translate("TradeOperation", u"dd/MM/yyyy", None))
    # retranslateUi




import logging
from PySide6.QtCore import Slot, Signal
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QWidget, QDataWidgetMapper
from PySide6.QtSql import QSqlTableModel
from jal.db.helpers import load_icon
from jal.db.db import JalModel


class AbstractOperationDetails(QWidget):
    dbUpdated = Signal()

    # ui_class should be a widget class generated by Qt-designer.
    # This widget should have QLabel main_label, and 2 QPushButtons: commit_button and revert_button
    def __init__(self, parent=None, ui_class=None):
        super().__init__(parent)
        assert ui_class is not None, "Can't create operation class without UI provided"
        self.ui = ui_class()
        self.ui.setupUi(self)

        self.model = None
        self.table_name = ''
        self.mapper = None
        self.modified = False
        self.name = self.name = self.ui.main_label.text()
        self.operation_type = None

        self.bold_font = QFont()
        self.bold_font.setBold(True)

        self.ui.commit_button.setIcon(load_icon("accept.png"))
        self.ui.revert_button.setIcon(load_icon("cancel.png"))

    def _init_db(self, table_name):
        self.table_name = table_name
        self.model = JalModel(self, table_name)
        self.model.setEditStrategy(QSqlTableModel.OnManualSubmit)

        self.mapper = QDataWidgetMapper(self.model)
        self.mapper.setModel(self.model)
        self.mapper.setSubmitPolicy(QDataWidgetMapper.AutoSubmit)

        self.model.dataChanged.connect(self.onDataChange)
        self.ui.commit_button.clicked.connect(self.saveChanges)
        self.ui.revert_button.clicked.connect(self.revertChanges)

    def set_id(self, oid):
        self.model.setFilter(f"id={oid}")
        self.mapper.setCurrentModelIndex(self.model.index(0, 0))

    @Slot()
    def onDataChange(self, _index_start, _index_stop, _role):
        self.modified = True
        self.ui.commit_button.setEnabled(True)
        self.ui.revert_button.setEnabled(True)

    @Slot()
    def saveChanges(self):
        if self._validated():
            self._save()

    def _validated(self):   # May be used in descendant classes
        return True

    def _save(self):
        if not self.model.submitAll():
            logging.fatal(self.tr("Operation submit failed: ") + self.model.lastError().text())
            return False
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        if self.modified:
            self.revertChanges()
            logging.warning(self.tr("Unsaved changes were reverted to create new operation"))
        self.model.setFilter(f"{self.table_name}.id = 0")
        new_record = self.prepareNew(account_id)
        assert self.model.insertRows(0, 1)
        self.model.setRecord(0, new_record)
        self.mapper.toLast()

    def prepareNew(self, account_id):
        new_record = self.model.record()
        new_record.setNull("id")
        new_record.setValue("op_type", self.operation_type)
        return new_record

    def copyNew(self):
        row = self.mapper.currentIndex()
        new_record = self.copyToNew(row)
        self.model.setFilter(f"{self.table_name}.id = 0")
        assert self.model.insertRows(0, 1)
        self.model.setRecord(0, new_record)
        self.mapper.toLast()

    def copyToNew(self, row):
        new_record = self.model.record(row)
        return new_record




from PySide6.QtCore import QObject, Signal, QDateTime
from PySide6.QtWidgets import QApplication


#-----------------------------------------------------------------------------------------------------------------------
# Base parent class for various API used by JAL for downloading information about purchases (slips)
class ReceiptAPI(QObject):
    slip_load_failed = Signal()
    slip_load_ok = Signal()

    def __init__(self):
        super().__init__()

    def tr(self, text):
        return QApplication.translate("ReceiptAPI", text)

    # Provides a list of parameters required for slip query if manual input is in use in form of dictionary
    # { "parameter_name" : "parameter_type" }
    @staticmethod
    def parameters_list() -> dict:
        raise NotImplementedError(f"parameters_list() shouldn't be called for ReceiptAPI class")

    # Method performs required actions to have active API session that may be used for queries
    # Returns True after successful activation and False otherwise
    def activate_session(self) -> bool:
        raise NotImplementedError(f"activate_session() method is not implemented in {type(self).__name__}")

    # Request slip data via API
    def query_slip(self):
        raise NotImplementedError(f"query_slip() method is not implemented in {type(self).__name__}")

    # Returns a list of purchased items (as dictionaries with name, price, amount, etc)
    def slip_lines(self) -> list:
        raise NotImplementedError(f"slip_lines() method is not implemented in {type(self).__name__}")

    # Returns a shop name where purchase was done (if it is possible to get)
    def shop_name(self) -> str:
        raise NotImplementedError(f"shop_name() method is not implemented in {type(self).__name__}")

    # Returns data/time of the operation from the receipt
    def datetime(self) -> QDateTime:
        raise NotImplementedError(f"datetime() method is not implemented in {type(self).__name__}")



import logging
from datetime import datetime
from dateutil import tz
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QStringListModel, QByteArray
from PySide6.QtWidgets import QHeaderView
from PySide6.QtSql import QSqlTableModel
from PySide6.QtGui import QFont
from jal.ui.widgets.ui_corporate_action_operation import Ui_CorporateActionOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.delegates import WidgetMapperDelegateBase, AssetSelectorDelegate, FloatDelegate
from jal.db.view_model import JalViewModel
from jal.db.helpers import load_icon, localize_decimal
from jal.db.operations import LedgerTransaction


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate,
                          'asset_id': self.symbol_delegate,
                          'qty': self.decimal_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_CorporateActionOperation)
        self.operation_type = LedgerTransaction.CorporateAction
        self.combo_model = None

        self.asset_delegate = AssetSelectorDelegate()
        self.float_delegate = FloatDelegate(2)
        self.percent_delegate = FloatDelegate(2, percent=True)

        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.add_button.setIcon(load_icon("add.png"))
        self.ui.del_button.setIcon(load_icon("remove.png"))
        self.ui.results_table.horizontalHeader().setFont(self.bold_font)
        self.ui.arrow.setText(" 🡆 ")  # it crashes if added via Qt-Designer

        self.ui.add_button.clicked.connect(self.addResult)
        self.ui.del_button.clicked.connect(self.delResult)

        super()._init_db("asset_actions")
        self.combo_model = QStringListModel([self.tr("N/A"),
                                             self.tr("Merger"),
                                             self.tr("Spin-Off"),
                                             self.tr("Symbol change"),
                                             self.tr("Split"),
                                             self.tr("Delisting")])
        self.ui.type.setModel(self.combo_model)

        self.mapper.setItemDelegate(CorporateActionWidgetDelegate(self.mapper))

        self.results_model = ResultsModel(self.ui.results_table, "action_results")
        self.results_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.ui.results_table.setModel(self.results_model)

        self.results_model.dataChanged.connect(self.onDataChange)
        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.asset_widget.changed.connect(self.mapper.submit)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.asset_widget, self.model.fieldIndex("asset_id"))
        self.mapper.addMapping(self.ui.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.ui.qty_edit, self.model.fieldIndex("qty"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))
        self.mapper.addMapping(self.ui.type, self.model.fieldIndex("type"), QByteArray().setRawData("currentIndex", 12))

        self.ui.results_table.setItemDelegateForColumn(2, self.asset_delegate)
        self.ui.results_table.setItemDelegateForColumn(3, self.float_delegate)
        self.ui.results_table.setItemDelegateForColumn(4, self.percent_delegate)

        self.model.select()
        self.results_model.select()
        self.results_model.configureView()

    def set_id(self, oid):
        self.results_model.setFilter(f"action_results.action_id = {oid}")  # First we need to select right children
        super().set_id(oid)

    @Slot()
    def addResult(self):
        new_record = self.results_model.record()
        new_record.setValue("qty", 0)
        new_record.setValue("value_share", 0)
        if not self.results_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.results_model.lastError().text())
            return

    @Slot()
    def delResult(self):
        selection = self.ui.results_table.selectionModel().selection().indexes()
        for idx in selection:
            self.results_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    def _save(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            oid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if oid is None:  # we just have saved new action record and need last inserted id
                oid = self.model.last_insert_id()
            for row in range(self.results_model.rowCount()):
                self.results_model.setData(self.results_model.index(row, self.results_model.fieldIndex("action_id")), oid)
            if not self.results_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.results_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.results_model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.results_model.setFilter(f"action_results.action_id = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        new_record.setValue("account_id", account_id)
        new_record.setValue("type", 0)
        new_record.setValue("asset_id", 0)
        new_record.setValue("qty", '0')
        new_record.setValue("note", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.results_model.rowCount()):
            child_records.append(self.results_model.record(row))
        self.results_model.setFilter(f"action_results.action_id = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("action_id")
            assert self.results_model.insertRows(0, 1)
            self.results_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        return new_record


# ----------------------------------------------------------------------------------------------------------------------
class ResultsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "action_id", self.tr("Asset"), self.tr("Qty"), self.tr("Share, %")]

    def footerData(self, section, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 2:
                return self.tr("Total")
            elif section == 4:
                total = Decimal('0')
                for row in range(self.rowCount()):
                    try:
                        value = Decimal(self.index(row, section).data())
                    except:
                        value = Decimal('0')
                    total += value
                return localize_decimal(total, precision=2, percent=True)
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 2:
                return Qt.AlignLeft | Qt.AlignVCenter
            else:
                return Qt.AlignRight | Qt.AlignVCenter
        return None

    def configureView(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 100)
        self._view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'transfer_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QDateTimeEdit, QGridLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QWidget)

from jal.widgets.account_select import AccountCurrencyLabel
from jal.widgets.reference_selector import (AccountSelector, AssetSelector)

class Ui_TransferOperation(object):
    def setupUi(self, TransferOperation):
        if not TransferOperation.objectName():
            TransferOperation.setObjectName(u"TransferOperation")
        TransferOperation.resize(968, 283)
        self.layout = QGridLayout(TransferOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.to_currency = AccountCurrencyLabel(TransferOperation)
        self.to_currency.setObjectName(u"to_currency")

        self.layout.addWidget(self.to_currency, 3, 7, 1, 1)

        self.from_currency = AccountCurrencyLabel(TransferOperation)
        self.from_currency.setObjectName(u"from_currency")

        self.layout.addWidget(self.from_currency, 2, 7, 1, 1)

        self.from_account_label = QLabel(TransferOperation)
        self.from_account_label.setObjectName(u"from_account_label")
        self.from_account_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.from_account_label, 2, 0, 1, 1)

        self.to_account_widget = AccountSelector(TransferOperation)
        self.to_account_widget.setObjectName(u"to_account_widget")

        self.layout.addWidget(self.to_account_widget, 3, 4, 1, 1)

        self.note_label = QLabel(TransferOperation)
        self.note_label.setObjectName(u"note_label")
        self.note_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.note_label, 7, 0, 1, 1)

        self.copy_amount_btn = QPushButton(TransferOperation)
        self.copy_amount_btn.setObjectName(u"copy_amount_btn")

        self.layout.addWidget(self.copy_amount_btn, 1, 6, 1, 1)

        self.deposit_timestamp = QDateTimeEdit(TransferOperation)
        self.deposit_timestamp.setObjectName(u"deposit_timestamp")
        self.deposit_timestamp.setCalendarPopup(True)
        self.deposit_timestamp.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.deposit_timestamp, 3, 2, 1, 2)

        self.main_label = QLabel(TransferOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 1)

        self.number_label = QLabel(TransferOperation)
        self.number_label.setObjectName(u"number_label")
        self.number_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.number_label, 1, 7, 1, 1)

        self.horizontal_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontal_spacer, 1, 10, 1, 1)

        self.fee_account_widget = AccountSelector(TransferOperation)
        self.fee_account_widget.setObjectName(u"fee_account_widget")

        self.layout.addWidget(self.fee_account_widget, 5, 4, 1, 1)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.vertical_spacer, 8, 0, 1, 1)

        self.withdrawal = QLineEdit(TransferOperation)
        self.withdrawal.setObjectName(u"withdrawal")
        self.withdrawal.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.withdrawal, 2, 5, 1, 2)

        self.note = QLineEdit(TransferOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 7, 2, 1, 8)

        self.commit_button = QPushButton(TransferOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 11, 1, 1)

        self.copy_date_btn = QPushButton(TransferOperation)
        self.copy_date_btn.setObjectName(u"copy_date_btn")

        self.layout.addWidget(self.copy_date_btn, 1, 3, 1, 1)

        self.withdrawal_timestamp = QDateTimeEdit(TransferOperation)
        self.withdrawal_timestamp.setObjectName(u"withdrawal_timestamp")
        self.withdrawal_timestamp.setCalendarPopup(True)
        self.withdrawal_timestamp.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.withdrawal_timestamp, 2, 2, 1, 2)

        self.number = QLineEdit(TransferOperation)
        self.number.setObjectName(u"number")

        self.layout.addWidget(self.number, 1, 9, 1, 1)

        self.from_account_widget = AccountSelector(TransferOperation)
        self.from_account_widget.setObjectName(u"from_account_widget")

        self.layout.addWidget(self.from_account_widget, 2, 4, 1, 1)

        self.fee_check = QCheckBox(TransferOperation)
        self.fee_check.setObjectName(u"fee_check")

        self.layout.addWidget(self.fee_check, 5, 2, 1, 2, Qt.AlignRight)

        self.to_account_label = QLabel(TransferOperation)
        self.to_account_label.setObjectName(u"to_account_label")
        self.to_account_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.to_account_label, 3, 0, 1, 1)

        self.date_label = QLabel(TransferOperation)
        self.date_label.setObjectName(u"date_label")

        self.layout.addWidget(self.date_label, 1, 2, 1, 1)

        self.revert_button = QPushButton(TransferOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 12, 1, 1)

        self.amount_label = QLabel(TransferOperation)
        self.amount_label.setObjectName(u"amount_label")

        self.layout.addWidget(self.amount_label, 1, 5, 1, 1)

        self.fee_currency = AccountCurrencyLabel(TransferOperation)
        self.fee_currency.setObjectName(u"fee_currency")

        self.layout.addWidget(self.fee_currency, 5, 7, 1, 1)

        self.asset_widget = AssetSelector(TransferOperation)
        self.asset_widget.setObjectName(u"asset_widget")

        self.layout.addWidget(self.asset_widget, 2, 9, 1, 1)

        self.asset_check = QCheckBox(TransferOperation)
        self.asset_check.setObjectName(u"asset_check")

        self.layout.addWidget(self.asset_check, 5, 9, 1, 1)

        self.account_label = QLabel(TransferOperation)
        self.account_label.setObjectName(u"account_label")

        self.layout.addWidget(self.account_label, 1, 4, 1, 1)

        self.fee = QLineEdit(TransferOperation)
        self.fee.setObjectName(u"fee")
        self.fee.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.fee, 5, 5, 1, 2)

        self.deposit = QLineEdit(TransferOperation)
        self.deposit.setObjectName(u"deposit")
        self.deposit.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.layout.addWidget(self.deposit, 3, 5, 1, 2)

        self.value_label = QLabel(TransferOperation)
        self.value_label.setObjectName(u"value_label")

        self.layout.addWidget(self.value_label, 3, 9, 1, 1)


        self.retranslateUi(TransferOperation)

        QMetaObject.connectSlotsByName(TransferOperation)
    # setupUi

    def retranslateUi(self, TransferOperation):
        TransferOperation.setWindowTitle(QCoreApplication.translate("TransferOperation", u"Form", None))
        self.to_currency.setText(QCoreApplication.translate("TransferOperation", u"CUR", None))
        self.from_currency.setText(QCoreApplication.translate("TransferOperation", u"CUR", None))
        self.from_account_label.setText(QCoreApplication.translate("TransferOperation", u"From", None))
        self.note_label.setText(QCoreApplication.translate("TransferOperation", u"Note", None))
#if QT_CONFIG(tooltip)
        self.copy_amount_btn.setToolTip(QCoreApplication.translate("TransferOperation", u"Copy value", None))
#endif // QT_CONFIG(tooltip)
        self.copy_amount_btn.setText(QCoreApplication.translate("TransferOperation", u"\u2193", None))
        self.deposit_timestamp.setDisplayFormat(QCoreApplication.translate("TransferOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.main_label.setText(QCoreApplication.translate("TransferOperation", u"Transfer", None))
        self.number_label.setText(QCoreApplication.translate("TransferOperation", u"#", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("TransferOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
#if QT_CONFIG(tooltip)
        self.copy_date_btn.setToolTip(QCoreApplication.translate("TransferOperation", u"Copy value", None))
#endif // QT_CONFIG(tooltip)
        self.copy_date_btn.setText(QCoreApplication.translate("TransferOperation", u"\u2193", None))
        self.withdrawal_timestamp.setDisplayFormat(QCoreApplication.translate("TransferOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.fee_check.setText(QCoreApplication.translate("TransferOperation", u"Include &fee", None))
        self.to_account_label.setText(QCoreApplication.translate("TransferOperation", u"To", None))
        self.date_label.setText(QCoreApplication.translate("TransferOperation", u"Date/Time", None))
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("TransferOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
        self.amount_label.setText(QCoreApplication.translate("TransferOperation", u"Amount", None))
        self.fee_currency.setText(QCoreApplication.translate("TransferOperation", u"CUR", None))
        self.asset_check.setText(QCoreApplication.translate("TransferOperation", u"Asset transfer", None))
        self.account_label.setText(QCoreApplication.translate("TransferOperation", u"Account", None))
        self.value_label.setText(QCoreApplication.translate("TransferOperation", u"(asset cost basis in new currency)", None))
    # retranslateUi




# Base classes for Category, Tag and Peer reports
from decimal import Decimal
from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from jal.db.asset import JalAsset
from jal.db.helpers import localize_decimal
from jal.db.operations_model import OperationsModel

class ReportOperationsModel(OperationsModel):
    def __init__(self, parent_view):
        self._total = Decimal('0')
        self._total_currency = 0
        self._total_currency_name = ''
        super().__init__(parent_view)

    # Is used by view to display footer Title, Total amount and Total currency with right font and alignment in columns 3-5
    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 4:
                return localize_decimal(self._total, precision=2)
            elif section == 5:
                return self._total_currency_name
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 3 or section == 4:
                return Qt.AlignRight | Qt.AlignVCenter
            else:
                return Qt.AlignLeft | Qt.AlignVCenter
        return None

    # Triggers view update if display parameters were changed
    def updateView(self, update: bool, dates_range: tuple, total_currency_id: int):
        if self._begin != dates_range[0]:
            self._begin = dates_range[0]
            update = True
        if self._end != dates_range[1]:
            self._end = dates_range[1]
            update = True
        if self._total_currency != total_currency_id:
            self._total_currency = total_currency_id
            self._total_currency_name = JalAsset(total_currency_id).symbol()
            update = True
        if update:
            self.prepareData()
            self.configureView()



from decimal import Decimal
from jal.db.db import JalDB
import jal.db.account
import jal.db.asset
import jal.db.operations
from jal.db.asset import JalAsset


class JalClosedTrade(JalDB):
    def __init__(self, id: int = 0) -> None:
        super().__init__()
        self._id = id
        self._data = self._read("SELECT account_id, asset_id, open_op_type, open_op_id, open_timestamp, open_price, "
                                "close_op_type, close_op_id, close_timestamp, close_price, qty "
                                "FROM trades_closed WHERE id=:id", [(":id", self._id)], named=True)
        if self._data:
            self._account = jal.db.account.JalAccount(self._data['account_id'])
            self._asset = jal.db.asset.JalAsset(self._data['asset_id'])
            self._open_op = jal.db.operations.LedgerTransaction.get_operation(self._data['open_op_type'], self._data['open_op_id'], jal.db.operations.Transfer.Incoming)
            self._close_op = jal.db.operations.LedgerTransaction.get_operation(self._data['close_op_type'], self._data['close_op_id'], jal.db.operations.Transfer.Outgoing)
            self._open_price = Decimal(self._data['open_price'])
            self._close_price = Decimal(self._data['close_price'])
            self._qty = Decimal(self._data['qty'])
        else:
            self._account = self._asset = self._open_op = self._close_op = None
            self._open_price = self._close_price = self._qty = Decimal('0')

    def dump(self) -> list:
        return [
            self._asset.symbol(self._account.currency()),
            self._open_op.timestamp(),
            self._close_op.timestamp(),
            self._open_price,
            self._close_price,
            self._qty,
            self.fee(),
            self.profit(),
            self.profit(percent=True)
        ]

    def id(self) -> int:
        return self._id

    def asset(self) -> jal.db.asset.JalAsset:
        return self._asset

    def symbol(self) -> str:
        return self._asset.symbol(self._account.currency())

    def open_operation(self):
        return self._open_op

    def close_operation(self):
        return self._close_op

    def qty(self) -> Decimal:
        return self._qty

    def open_price(self) -> Decimal:
        return self._open_price

    def close_price(self) -> Decimal:
        return self._close_price

    # If currency_id is different from trade account currency then adjusts value to the rate of currency for
    # given timestamp and returns it. Otherwise, simply returns unchanged value
    def adjusted(self, value: Decimal, currency_id: int, timestamp: int) -> Decimal:
        if currency_id and currency_id != self._account.currency():
            return value * JalAsset(self._account.currency()).quote(timestamp, currency_id)[1]
        else:
            return value

    # Returns opening amount of the trade (open price x qty)
    # If currency_id isn't 0 then converts amount into given currency using settlement date rate
    # If no_settlement is set to True then transaction date is used for conversion
    def open_amount(self, currency_id: int = 0, no_settlement=False) -> Decimal:
        timestamp = self._open_op.timestamp() if no_settlement else self._open_op.settlement()
        return self.adjusted(self._open_price * abs(self._qty), currency_id, timestamp)

    # Returns closing amount of the trade (close price x qty)
    # If currency_id isn't 0 then converts amount into given currency using settlement date rate
    # If no_settlement is set to True then transaction date is used for conversion
    def close_amount(self, currency_id: int = 0, no_settlement=False) -> Decimal:
        timestamp = self._close_op.timestamp() if no_settlement else self._close_op.settlement()
        return self.adjusted(self._close_price * abs(self._qty), currency_id, timestamp)

    # Fee of opening part of the deal
    # if currency_id isn't 0 then returns fee converted into given currency
    def open_fee(self, currency_id: int = 0) -> Decimal:
        if self._open_op.type() == jal.db.operations.LedgerTransaction.Trade:
            o_fee = self._open_op.fee() * abs(self._qty / self._open_op.qty())
            return self.adjusted(o_fee, currency_id, self._open_op.timestamp())
        else:
            return Decimal('0')

    # Fee of closing part of the deal
    # if currency_id isn't 0 then returns fee converted into given currency
    def close_fee(self, currency_id: int = 0) -> Decimal:
        if self._close_op.type() == jal.db.operations.LedgerTransaction.Trade:
            c_fee = self._close_op.fee() * abs(self._qty / self._close_op.qty())
            return self.adjusted(c_fee, currency_id, self._close_op.timestamp())
        else:
            return Decimal('0')

    # Total fee for the trade
    def fee(self, currency_id: int = 0) -> Decimal:
        return self.open_fee(currency_id) + self.close_fee(currency_id)

    def profit(self, percent=False) -> Decimal:
        profit = self._qty * (self._close_price - self._open_price) - self.fee()
        if percent:
            profit = Decimal('100') * profit / (self._qty * self._open_price) if self._open_price else Decimal('0')
        return profit



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'profit_loss_report.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QFrame, QHBoxLayout,
    QHeaderView, QLabel, QPushButton, QSizePolicy,
    QSpacerItem, QTableView, QVBoxLayout, QWidget)

from jal.widgets.custom.date_range_selector import DateRangeSelector
from jal.widgets.reference_selector import AccountSelector

class Ui_ProfitLossReportWidget(object):
    def setupUi(self, ProfitLossReportWidget):
        if not ProfitLossReportWidget.objectName():
            ProfitLossReportWidget.setObjectName(u"ProfitLossReportWidget")
        ProfitLossReportWidget.resize(648, 301)
        self.verticalLayout = QVBoxLayout(ProfitLossReportWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.ReportParamsFrame = QFrame(ProfitLossReportWidget)
        self.ReportParamsFrame.setObjectName(u"ReportParamsFrame")
        self.ReportParamsFrame.setFrameShape(QFrame.Panel)
        self.ReportParamsFrame.setFrameShadow(QFrame.Sunken)
        self.horizontalLayout = QHBoxLayout(self.ReportParamsFrame)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(2, 2, 2, 2)
        self.ReportRange = DateRangeSelector(self.ReportParamsFrame)
        self.ReportRange.setObjectName(u"ReportRange")
        self.ReportRange.setProperty("ItemsList", u"QTD;YTD;this_year;last_year")

        self.horizontalLayout.addWidget(self.ReportRange)

        self.ReportAccountLbl = QLabel(self.ReportParamsFrame)
        self.ReportAccountLbl.setObjectName(u"ReportAccountLbl")

        self.horizontalLayout.addWidget(self.ReportAccountLbl)

        self.ReportAccountEdit = AccountSelector(self.ReportParamsFrame)
        self.ReportAccountEdit.setObjectName(u"ReportAccountEdit")

        self.horizontalLayout.addWidget(self.ReportAccountEdit)

        self.CurrencyLbl = QLabel(self.ReportParamsFrame)
        self.CurrencyLbl.setObjectName(u"CurrencyLbl")

        self.horizontalLayout.addWidget(self.CurrencyLbl)

        self.ReportFrameSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.ReportFrameSpacer)

        self.SaveButton = QPushButton(self.ReportParamsFrame)
        self.SaveButton.setObjectName(u"SaveButton")

        self.horizontalLayout.addWidget(self.SaveButton)


        self.verticalLayout.addWidget(self.ReportParamsFrame)

        self.ReportTableView = QTableView(ProfitLossReportWidget)
        self.ReportTableView.setObjectName(u"ReportTableView")
        self.ReportTableView.setFrameShape(QFrame.Panel)
        self.ReportTableView.setFrameShadow(QFrame.Sunken)
        self.ReportTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.ReportTableView.setAlternatingRowColors(True)
        self.ReportTableView.setGridStyle(Qt.DotLine)
        self.ReportTableView.setWordWrap(False)
        self.ReportTableView.verticalHeader().setVisible(False)
        self.ReportTableView.verticalHeader().setMinimumSectionSize(20)
        self.ReportTableView.verticalHeader().setDefaultSectionSize(20)

        self.verticalLayout.addWidget(self.ReportTableView)


        self.retranslateUi(ProfitLossReportWidget)

        QMetaObject.connectSlotsByName(ProfitLossReportWidget)
    # setupUi

    def retranslateUi(self, ProfitLossReportWidget):
        ProfitLossReportWidget.setWindowTitle(QCoreApplication.translate("ProfitLossReportWidget", u"P&L", None))
        self.ReportAccountLbl.setText(QCoreApplication.translate("ProfitLossReportWidget", u"Account:", None))
        self.CurrencyLbl.setText(QCoreApplication.translate("ProfitLossReportWidget", u"Currency: ", None))
        self.SaveButton.setText(QCoreApplication.translate("ProfitLossReportWidget", u"Save...", None))
    # retranslateUi




import uuid
import logging
import json
import requests
import time
from urllib import parse
from decimal import Decimal
from PySide6.QtCore import Qt, Signal, Slot, QUrl, QDateTime, QTimeZone
from PySide6.QtWidgets import QDialog
from PySide6.QtWebEngineCore import QWebEngineUrlRequestInterceptor, QWebEngineProfile, QWebEnginePage
from jal.data_import.receipt_api.receipt_api import ReceiptAPI
from jal.db.settings import JalSettings
from jal.net.helpers import get_web_data, post_web_data
from jal.ui.ui_login_fns_dlg import Ui_LoginFNSDialog


#-----------------------------------------------------------------------------------------------------------------------
class ReceiptRuFNS(ReceiptAPI):
    MAX_ATTEMPTS = 5
    timestamp_patterns = ['yyyyMMddTHHmm', 'yyyyMMddTHHmmss', 'yyyy-MM-ddTHH:mm', 'yyyy-MM-ddTHH:mm:ss']

    def __init__(self, qr_text='', aux_data='', params=None):
        super().__init__()
        self.session_id = ''
        self.slip_json = {}
        if params is None:
            try:
                params = parse.parse_qs(qr_text)
                self.date_time = ''
                for timestamp_pattern in self.timestamp_patterns:
                    datetime_value = QDateTime.fromString(params['t'][0], timestamp_pattern)
                    if datetime_value.isValid():
                        self.date_time = datetime_value.toString("yyyyMMddThhmmss")
                        break
                if not self.date_time:
                    raise ValueError(self.tr("FNS QR available but date/time pattern isn't recognized: " + qr_text))
                self.amount = float(params['s'][0])
                self.fn = params['fn'][0]
                self.fd = params['i'][0]
                self.fp = params['fp'][0]
                self.op_type = params['n'][0]
            except Exception:
                raise ValueError(self.tr("FNS QR available but pattern isn't recognized: " + qr_text))
        else:
            self.date_time = params['Дата/время'].toString("yyyyMMddThhmmss")
            self.fn = params['ФН']
            self.fd = params['ФД']
            self.fp = params['ФП']
            self.amount = float(params['Сумма'])
            self.op_type = params['Тип']
        self.web_session = requests.Session()
        self.web_session.headers['ClientVersion'] = '2.9.0'
        self.web_session.headers['Device-Id'] = str(uuid.uuid1())
        self.web_session.headers['Device-OS'] = 'Android'
        self.web_session.headers['Content-Type'] = 'application/json; charset=UTF-8'
        self.web_session.headers['Accept-Encoding'] = 'gzip'
        self.web_session.headers['User-Agent'] = 'okhttp/4.2.2'

    @staticmethod
    def parameters_list() -> dict:
        parameters = {
            "Дата/время": QDateTime.currentDateTime(QTimeZone.UTC),
            "ФН": '',
            "ФД": '',
            "ФП": '',
            "Сумма": Decimal('0'),
            "Тип": {1: "Покупка", 2: "Возврат"}
        }
        return parameters

    def activate_session(self) -> bool:
        self.session_id = JalSettings().getValue('RuTaxSessionId', default='')
        if not self.session_id:
            self.__do_login()
        if not self.session_id:
            logging.warning(self.tr("No FNS SessionId available"))
            return False
        self.web_session.headers['sessionId'] = self.session_id
        response = self.web_session.get('https://irkkt-mobile.nalog.ru:8888/v2/tickets')  # Just to check authorization status
        if response.status_code == 200:
            return True
        if response.status_code == 401:
            logging.info(self.tr("Unauthorized with reason: ") + f"{response.text}")
            return self.__refresh_session()
        else:
            logging.error(self.tr("FNS API failed with: ") + f"{response.status_code}/{response.text}")
            return False

    def __refresh_session(self) -> bool:
        if not self.session_id:
            return False
        logging.info(self.tr("Refreshing FNS session..."))
        client_secret = JalSettings().getValue('RuTaxClientSecret')
        refresh_token = JalSettings().getValue('RuTaxRefreshToken')
        self.web_session.headers['sessionId'] = self.session_id
        payload = '{' + f'"client_secret":"{client_secret}","refresh_token":"{refresh_token}"' + '}'
        response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/mobile/users/refresh', data=payload)
        if response.status_code == 200:
            logging.info(self.tr("FNS session refreshed: ") + f"{response.text}")
            json_content = json.loads(response.text)
            self.session_id = json_content['sessionId']
            new_refresh_token = json_content['refresh_token']
            settings = JalSettings()
            settings.setValue('RuTaxSessionId', self.session_id)
            settings.setValue('RuTaxRefreshToken', new_refresh_token)
            self.web_session.headers['sessionId'] = self.session_id
            return True
        else:
            logging.error(self.tr("Can't refresh FNS session, response: ") + f"{response.status_code}/{response.text}")
            JalSettings().setValue('RuTaxSessionId', '')
            self.session_id = ''
            return False

    def __do_login(self):
        login_dialog = LoginFNS()
        if login_dialog.exec() == QDialog.Accepted:
            self.session_id = JalSettings().getValue('RuTaxSessionId')

    def query_slip(self):
        json_content = {}
        payload = '{' + f'"qr": "t={self.date_time}&s={self.amount:.2f}&fn={self.fn}&i={self.fd}&fp={self.fp}&n={self.op_type}"' + '}'
        for i in range(self.MAX_ATTEMPTS):
            response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/ticket', data=payload)
            if response.status_code != 200:
                logging.error(self.tr("Get ticket id failed: ") + f"{response.status_code}/{response.text} for {payload}")
                return
            logging.info(self.tr("Receipt found: " + response.text))
            json_content = json.loads(response.text)
            if json_content['status'] == 2:  # Valid slip status is 2, other statuses are not fully clear
                break
            logging.warning(self.tr("Operation might be pending on server side. Trying again."))
            time.sleep(0.5)  # wait half a second before next attempt
        if json_content['status'] == 2:
            url = "https://irkkt-mobile.nalog.ru:8888/v2/tickets/" + json_content['id']
            response = self.web_session.get(url)
            if response.status_code != 200:
                logging.error(self.tr("Receipt load failed: ") + f"{response}/{response.text}")
                self.slip_load_failed.emit()
            logging.info(self.tr("Receipt was loaded: " + response.text))
            self.slip_json = self.__slip_data(json.loads(response.text))
            self.slip_load_ok.emit()
        else:
            self.slip_load_failed.emit()

    # Slip data might be in a root element or in ticket/document/receipt
    # This method returns actual slip as result
    def __slip_data(self, json_data) -> dict:
        if 'ticket' in json_data:
            sub = json_data['ticket']
            if 'document' in sub:
                sub = sub['document']
                if 'receipt' in sub:
                    receipt = sub['receipt']
                else:
                    logging.error(self.tr("Can't find 'receipt' tag in json 'document' from FNS"))
                    return {}
            else:
                logging.error(self.tr("Can't find 'document' tag in json 'ticket' from FNS"))
                return {}
        else:
            receipt = json_data
        return receipt

    # ----------------------------------------------------------------------------------------------------------------
    # Gets company name by Russian INN
    # Returns short or long name if fount and initial INN otherwise
    def shop_name(self) -> str:
        if 'user' in self.slip_json:
            return self.slip_json['user']
        if 'userInn' in self.slip_json:
            inn = self.slip_json['userInn']
        else:
            return ''
        if len(inn) != 10 and len(inn) != 12:
            logging.warning(self.tr("Incorrect length of INN. Can't get company name."))
            return inn
        region_list = "77,78,01,02,03,04,05,06,07,08,09,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,"\
                      "30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,"\
                      "61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,79,83,86,87,89,91,92,99"
        params = {'vyp3CaptchaToken': '', 'page': '', 'query': inn, 'region': region_list,
                  'PreventChromeAutocomplete': ''}
        token_data = json.loads(post_web_data('https://egrul.nalog.ru/', params))
        if 't' not in token_data:
            return inn
        result = json.loads(get_web_data('https://egrul.nalog.ru/search-result/' + token_data['t']))
        try:
            return result['rows'][0]['c']   # Return short name if exists
        except:
            pass
        try:
            return result['rows'][0]['n']   # Return long name if exists
        except:
            logging.warning(self.tr("Can't get company name from: ") + result)
            return inn

    def datetime(self) -> QDateTime:
        if 'dateTime' in self.slip_json:
            receipt_datetime = QDateTime.fromSecsSinceEpoch(int(self.slip_json['dateTime']))
        else:
            receipt_datetime = QDateTime()
        receipt_datetime.setTimeSpec(Qt.UTC)
        return receipt_datetime

    def slip_lines(self) -> list:
        PURCHASE = 1
        RETURN = 2

        if 'operationType' in self.slip_json:
            operation = int(self.slip_json['operationType'])
        else:
            logging.error(self.tr("Can't find 'operationType' tag in json 'ticket'"))
            return []
        lines = self.slip_json['items']
        for line in lines:
            line['unit_price'] = line.pop('price') / 100
            sign = -1 if operation == PURCHASE else +1
            line['amount'] = sign * line.pop('sum') / 100
            if line['quantity'] != 1:
                line['name'] = f"{line['name']} ({line['quantity']:g} x {line['unit_price']:.2f})"
        return lines


#-----------------------------------------------------------------------------------------------------------------------
class LoginFNS(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_LoginFNSDialog()
        self.ui.setupUi(self)

        self.phone_number = ''
        self.web_session = requests.Session()
        self.web_session.headers['ClientVersion'] = '2.9.0'
        self.web_session.headers['Device-Id'] = str(uuid.uuid1())
        self.web_session.headers['Device-OS'] = 'Android'
        self.web_session.headers['Content-Type'] = 'application/json; charset=UTF-8'
        self.web_session.headers['Accept-Encoding'] = 'gzip'
        self.web_session.headers['User-Agent'] = 'okhttp/4.2.2'
        self.web_profile = QWebEngineProfile(self)
        self.web_interceptor = RequestInterceptor(self)
        self.web_interceptor.response_intercepted.connect(self.response_esia)
        self.web_profile.setUrlRequestInterceptor(self.web_interceptor)
        self.ui.ESIAWebView.setPage(QWebEnginePage(self.web_profile, self))

        self.ui.LoginMethodTabs.currentChanged.connect(self.on_tab_changed)
        self.ui.GetCodeBtn.clicked.connect(self.send_sms)
        self.ui.SMSLoginBtn.clicked.connect(self.login_sms)
        self.ui.FNSLoginBtn.clicked.connect(self.login_fns)

    def on_tab_changed(self, index):
        if index == 2:  # ESIA login selected
            self.login_esia()

    def send_sms(self):
        client_secret = JalSettings().getValue('RuTaxClientSecret')
        self.phone_number = self.ui.PhoneNumberEdit.text().replace('-', '')

        payload = '{' + f'"client_secret":"{client_secret}","phone":"{self.phone_number}"' + '}'
        response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/auth/phone/request', data=payload)
        if response.status_code != 204:
            logging.error(self.tr("FNS login failed: ") + f"{response}/{response.text}")
        else:
            logging.info(self.tr("SMS was requested successfully"))

    def login_sms(self):
        if not self.phone_number:
            return
        client_secret = JalSettings().getValue('RuTaxClientSecret')
        code = self.ui.CodeEdit.text()

        payload = '{' + f'"client_secret":"{client_secret}","code":"{code}","phone":"{self.phone_number}"' + '}'
        response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/auth/phone/verify', data=payload)
        if response.status_code != 200:
            logging.error(self.tr("FNS login failed: ") + f"{response.status_code}/{response.text}")
            return
        logging.info(self.tr("FNS login successful: ") + f"{response.text}")
        json_content = json.loads(response.text)
        new_session_id = json_content['sessionId']
        new_refresh_token = json_content['refresh_token']
        settings = JalSettings()
        settings.setValue('RuTaxSessionId', new_session_id)
        settings.setValue('RuTaxRefreshToken', new_refresh_token)
        self.accept()

    def login_fns(self):
        client_secret = JalSettings().getValue('RuTaxClientSecret')
        inn = self.ui.InnEdit.text()
        password = self.ui.PasswordEdit.text()

        payload = '{' + f'"client_secret":"{client_secret}","inn":"{inn}","password":"{password}"' + '}'
        response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/mobile/users/lkfl/auth', data=payload)
        if response.status_code != 200:
            logging.error(self.tr("FNS login failed: ") + f"{response.status_code}/{response.text}")
            return
        logging.info(self.tr("FNS login successful: ") + f"{response.text}")
        json_content = json.loads(response.text)
        new_session_id = json_content['sessionId']
        new_refresh_token = json_content['refresh_token']
        settings = JalSettings()
        settings.setValue('RuTaxSessionId', new_session_id)
        settings.setValue('RuTaxRefreshToken', new_refresh_token)
        self.accept()

    def login_esia(self):
        response = self.web_session.get('https://irkkt-mobile.nalog.ru:8888/v2/mobile/users/esia/auth/url')
        if response.status_code != 200:
            logging.error(self.tr("Get ESIA URL failed: ") + f"{response.status_code}/{response.text}")
            return
        json_content = json.loads(response.text)
        auth_url = json_content['url']
        self.ui.ESIAWebView.load(QUrl(auth_url))

    @Slot()
    def response_esia(self, auth_code, state):
        client_secret = JalSettings().getValue('RuTaxClientSecret')
        payload = '{' + f'"authorization_code": "{auth_code}", "client_secret": "{client_secret}", "state": "{state}"' \
                  + '}'
        response = self.web_session.post('https://irkkt-mobile.nalog.ru:8888/v2/mobile/users/esia/auth', data=payload)
        if response.status_code != 200:
            logging.error(self.tr("ESIA login failed: ") + f"{response.status_code}/{response.text}")
            return
        logging.info(self.tr("ESIA login successful: ") + f"{response.text}")
        json_content = json.loads(response.text)
        new_session_id = json_content['sessionId']
        new_refresh_token = json_content['refresh_token']
        settings = JalSettings()
        settings.setValue('RuTaxSessionId', new_session_id)
        settings.setValue('RuTaxRefreshToken', new_refresh_token)
        self.accept()


#-----------------------------------------------------------------------------------------------------------------------
class RequestInterceptor(QWebEngineUrlRequestInterceptor):
    response_intercepted = Signal(str, str)

    def __init__(self, parent = None):
        super().__init__(parent)
        self.session = None

    # At successful login ESIA page will give Response '302 Found' with URL to irkkt-mobile.nalog.ru which
    # contains authorization results. We can't intercept response so we block next request and
    # get parameters from it. Then communicate it via 'response_intercepted' signal
    def interceptRequest(self, info):
        url = info.firstPartyUrl().url()    # Get intercepted URL
        if str.startswith(url, "https://irkkt-mobile.nalog.ru:8888/"):
            info.block(True)
            params = dict(parse.parse_qsl(parse.urlsplit(url).query))
            auth_code = params['code']
            auth_state = params['state']
            logging.info(self.tr("ESIA login completed"))
            self.response_intercepted.emit(auth_code, auth_state)



import logging
from datetime import datetime
from dateutil import tz
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QByteArray
from PySide6.QtWidgets import QMessageBox, QHeaderView
from PySide6.QtSql import QSqlTableModel
from PySide6.QtGui import QFont
from jal.ui.widgets.ui_income_spending_operation import Ui_IncomeSpendingOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.db.view_model import JalViewModel
from jal.db.helpers import load_icon, localize_decimal, db_row2dict
from jal.db.operations import LedgerTransaction
from jal.widgets.delegates import WidgetMapperDelegateBase, FloatDelegate, CategorySelectorDelegate, TagSelectorDelegate


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_IncomeSpendingOperation)
        self.operation_type = LedgerTransaction.IncomeSpending
        super()._init_db("actions")

        self.category_delegate = CategorySelectorDelegate()
        self.tag_delegate = TagSelectorDelegate()
        self.float_delegate = FloatDelegate(2)

        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.a_currency.setText(self.tr("Paid in foreign currency:"))
        self.ui.details_table.horizontalHeader().setFont(self.bold_font)
        self.ui.add_button.setIcon(load_icon("add.png"))
        self.ui.del_button.setIcon(load_icon("remove.png"))
        self.ui.copy_button.setIcon(load_icon("copy.png"))

        self.ui.add_button.clicked.connect(self.add_child)
        self.ui.copy_button.clicked.connect(self.copy_child)
        self.ui.del_button.clicked.connect(self.delete_child)
        self.model.beforeInsert.connect(self.before_record_insert)
        self.model.beforeUpdate.connect(self.before_record_update)

        self.mapper.setItemDelegate(IncomeSpendingWidgetDelegate(self.mapper))

        self.details_model = DetailsModel(self.ui.details_table, "action_details")
        self.details_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.ui.details_table.setModel(self.details_model)
        self.details_model.dataChanged.connect(self.onDataChange)

        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.peer_widget.changed.connect(self.mapper.submit)
        self.ui.a_currency.changed.connect(self.mapper.submit)
        self.ui.a_currency.name_updated.connect(self.details_model.set_alternative_currency)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.currency, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.peer_widget, self.model.fieldIndex("peer_id"))
        self.mapper.addMapping(self.ui.a_currency, self.model.fieldIndex("alt_currency_id"), QByteArray("currency_id_str"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.ui.details_table.setItemDelegateForColumn(2, self.category_delegate)
        self.ui.details_table.setItemDelegateForColumn(3, self.tag_delegate)
        self.ui.details_table.setItemDelegateForColumn(4, self.float_delegate)
        self.ui.details_table.setItemDelegateForColumn(5, self.float_delegate)

        self.model.select()
        self.details_model.select()
        self.details_model.configure_view()

    def set_id(self, oid):
        self.details_model.setFilter(f"action_details.pid = {oid}")  # First we need to select right children
        super().set_id(oid)

    @Slot()
    def add_child(self):
        new_record = self.details_model.record()
        new_record.setNull("tag_id")
        new_record.setValue("amount", '0')
        new_record.setValue("amount_alt", '0')
        if not self.details_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.details_model.lastError().text())
            return

    @Slot()
    def copy_child(self):
        idx = self.ui.details_table.selectionModel().selection().indexes()
        src_record = self.details_model.record(idx[0].row())
        new_record = self.details_model.record()
        new_record.setValue("category_id", src_record.value("category_id"))
        if src_record.value("tag_id"):
            new_record.setValue("tag_id", src_record.value("tag_id"))
        else:
            new_record.setNull("tag_id")
        new_record.setValue("amount", src_record.value("amount"))
        new_record.setValue("amount_alt", src_record.value("amount_alt"))
        new_record.setValue("note", src_record.value("note"))
        if not self.details_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.details_model.lastError().text())
            return

    @Slot()
    def delete_child(self):
        selection = self.ui.details_table.selectionModel().selection().indexes()
        for idx in selection:
            self.details_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    def _validated(self):
        fields = db_row2dict(self.model, 0)
        if not fields['account_id'] or not fields['peer_id']:
            return False
        if not self.details_model.rowCount():
            QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Operation contains no details"), QMessageBox.Ok)
            return False
        for row in range(self.details_model.rowCount()):
            fields = db_row2dict(self.details_model, row)
            if fields['category_id'] is None or fields['category_id'] == 0:
                QMessageBox().warning(self, self.tr("Incomplete data"),
                                      self.tr("Category isn't set for '{}' (Amount: {})".format(fields['note'], fields['amount'])), QMessageBox.Ok)
                return False
        return True

    def _save(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            pid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if pid is None:  # we just have saved new action record and need last inserted id
                pid = self.model.last_insert_id()
            for row in range(self.details_model.rowCount()):   # Set PID for all child records
                self.details_model.setData(self.details_model.index(row, self.details_model.fieldIndex("pid")), pid)
            if not self.details_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.details_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.model.database().commit()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.details_model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.details_model.setFilter(f"action_details.pid = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("account_id", account_id)
        new_record.setValue("peer_id", 0)
        new_record.setValue("alt_currency_id", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.details_model.rowCount()):
            child_records.append(self.details_model.record(row))
        self.details_model.setFilter(f"action_details.pid = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("pid")
            if not record.value("tag_id"):
                record.setNull("tag_id")
            assert self.details_model.insertRows(0, 1)
            self.details_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        return new_record

    def before_record_insert(self, record):
        if record.value("alt_currency_id") == 0 or record.value("alt_currency_id") == '':
            record.setNull("alt_currency_id")

    def before_record_update(self, _row, record):
        self.before_record_insert(record)   # processing is the same as before insert


# ----------------------------------------------------------------------------------------------------------------------
class DetailsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "pid", self.tr("Category"), self.tr("Tag"),
                         self.tr("Amount"), self.tr("Amount"), self.tr("Note")]
        self.alt_currency_name = ''

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if section == 5 and orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section] + ', ' + self.alt_currency_name
        else:
            return super().headerData(section, orientation, role)

    def footerData(self, section, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 6:
                return self.tr("Total")
            elif section == 4 or section == 5:
                total = Decimal('0')
                for row in range(self.rowCount()):
                    try:
                        value = Decimal(self.index(row, section).data())
                    except:
                        value = Decimal('0')
                    total += value
                return localize_decimal(total, precision=2)
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 6:
                return Qt.AlignLeft | Qt.AlignVCenter
            else:
                return Qt.AlignRight | Qt.AlignVCenter
        return None

    def configure_view(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnHidden(5, True)
        self._view.setColumnWidth(2, 200)
        self._view.setColumnWidth(3, 200)
        self._view.setColumnWidth(4, 100)
        self._view.setColumnWidth(5, 100)
        self._view.horizontalHeader().setSectionResizeMode(6, QHeaderView.Stretch)
        self._view.horizontalHeader().moveSection(6, 0)

    def set_alternative_currency(self, currency_name):
        if currency_name:
            self._view.setColumnHidden(5, False)
            self.alt_currency_name = currency_name
            self.headerDataChanged.emit(Qt.Horizontal, 5, 5)
        else:
            for row in range(self.rowCount()):
                self.setData(self.index(row, self.fieldIndex("amount_alt")), '0')  # Reset all alternative amounts
            self._view.setColumnHidden(5, True)



from decimal import Decimal
from PySide6.QtCore import Qt, Slot, QObject
from jal.db.ledger import Ledger
from jal.db.asset import JalAsset
from jal.db.category import JalCategory
from jal.reports.reports import Reports
from jal.reports.operations_base import ReportOperationsModel
from jal.db.operations import LedgerTransaction
from jal.ui.reports.ui_category_report import Ui_CategoryReportWidget
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "CategoryReport"


# ----------------------------------------------------------------------------------------------------------------------
class CategoryOperationsModel(ReportOperationsModel):
    def __init__(self, parent_view):
        self._category_id = 0
        super().__init__(parent_view)

    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and section == 3:
            return self.tr("Total with peer ") + f"'{JalCategory(self._category_id).name()}':"
        return super().footerData(section, role)

    def updateView(self, category_id: int, dates_range: tuple, total_currency_id: int):
        update = False
        if self._category_id != category_id:
            self._category_id = category_id
            update = True
        super().updateView(update, dates_range, total_currency_id)

    def prepareData(self):
        self._data = []
        self._total = Decimal('0')
        self._data = Ledger.get_operations_by_category(self._begin, self._end, self._category_id)
        operations = [LedgerTransaction().get_operation(x['op_type'], x['id'], x['subtype']) for x in self._data]
        # Take only Income/Spending data as we expect Asset operations to be not relevant for this kind of report
        operations = [x for x in operations if x.type() == LedgerTransaction.IncomeSpending]
        for op in operations:
            op_amount = Decimal('0')
            for line in op.lines():
                if line['category_id'] == self._category_id:
                    try:
                        op_amount += Decimal(line['amount'])
                    except:
                        pass
            account_currency = op.account().currency()
            if account_currency == self._total_currency:
                self._total += op_amount
            else:
                rate = JalAsset(account_currency).quote(op.timestamp(), self._total_currency)[1]
                self._total += op_amount * rate
        self.modelReset.emit()


# ----------------------------------------------------------------------------------------------------------------------
class CategoryReport(QObject):
    def __init__(self):
        super().__init__()
        self.group = self.tr("Operations")
        self.name = self.tr("by Category")
        self.window_class = "CategoryReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class CategoryReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_CategoryReportWidget()
        self.ui.setupUi(self)
        self._parent = parent

        self.category_model = CategoryOperationsModel(self.ui.ReportTableView)
        self.ui.ReportTableView.setModel(self.category_model)
        self.ui.TotalCurrencyCombo.setIndex(JalAsset.get_base_currency())
        self.category_model.configureView()

        if settings is not None:
            self.ui.ReportRange.setRange(settings['begin_ts'], settings['end_ts'])
            self.ui.ReportCategoryEdit.selected_id = settings['category_id']
        self.connect_signals_and_slots()
        self.updateReport()

    def connect_signals_and_slots(self):
        self.ui.ReportRange.changed.connect(self.updateReport)
        self.ui.ReportCategoryEdit.changed.connect(self.updateReport)
        self.ui.TotalCurrencyCombo.changed.connect(self.updateReport)
        self.ui.ReportTableView.selectionModel().selectionChanged.connect(self.onOperationSelect)

    @Slot()
    def updateReport(self):
        self.ui.ReportTableView.model().updateView(
            category_id=self.ui.ReportCategoryEdit.selected_id, dates_range=self.ui.ReportRange.getRange(),
            total_currency_id=self.ui.TotalCurrencyCombo.selected_id)

    @Slot()
    def onOperationSelect(self, selected, _deselected):
        idx = selected.indexes()
        if idx:
            selected_row = idx[0].row()
            operation_type, operation_id = self.ui.ReportTableView.model().get_operation(selected_row)
            self.ui.OperationDetails.show_operation(operation_type, operation_id)
        else:
            self.ui.OperationDetails.show_operation(LedgerTransaction.NA, 0)



import logging
import pandas as pd
from decimal import Decimal
from PySide6.QtCore import Qt, Slot, QAbstractTableModel, QDateTime, QDate, QTime, QLocale
from PySide6.QtWidgets import QDialog, QHeaderView, QStyledItemDelegate, QLineEdit, QComboBox
from jal.widgets.reference_selector import CategorySelector, TagSelector
from jal.widgets.delegates import DateTimeEditWithReset
from jal.constants import CustomColor
from jal.widgets.helpers import dependency_present
from jal.db.helpers import localize_decimal, delocalize_decimal
from jal.db.peer import JalPeer
from jal.db.category import JalCategory
from jal.db.operations import LedgerTransaction
from jal.widgets.qr_scanner import ScanDialog
from jal.ui.ui_receipt_import_dlg import Ui_ImportShopReceiptDlg
from jal.data_import.category_recognizer import recognize_categories
from jal.data_import.receipt_api.receipts import ReceiptAPIFactory


DEFAULT_DATA_ROLE = Qt.UserRole + 1
#-----------------------------------------------------------------------------------------------------------------------
# Custom model to display and edit slip lines
class PandasLinesModel(QAbstractTableModel):
    def __init__(self, data, parent=None):
        super().__init__(parent)
        self._data = data

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        return super().flags(index) | Qt.ItemIsEditable

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                return self._data.iloc[index.row(), index.column()]
        return None

    def setData(self, index, value, role=Qt.EditRole):
        if index.isValid():
            if role == Qt.EditRole:
                self._data.iloc[index.row(), index.column()] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if col == 0:
                return self.tr("Product name")
            if col == 1:
                return self.tr("Category")
            if col == 3:
                return self.tr("Tag")
            if col == 4:
                return self.tr("Amount")
        return None


class SlipLinesDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def paint(self, painter, option, index):
        painter.save()
        pen = painter.pen()
        model = index.model()
        if index.column() == 0:
            text = model.data(index, Qt.DisplayRole)
            painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        if index.column() == 1:
            text = JalCategory(int(model.data(index, Qt.DisplayRole))).name()
            confidence = model.data(index.siblingAtColumn(2), Qt.DisplayRole)
            if confidence > 0.75:
                painter.fillRect(option.rect, CustomColor.LightGreen)
            elif confidence > 0.5:
                painter.fillRect(option.rect, CustomColor.LightYellow)
            else:
                painter.fillRect(option.rect, CustomColor.LightRed)
            painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        elif index.column() == 4:
            amount = model.data(index, Qt.DisplayRole)
            if amount == 2:
                pen.setColor(CustomColor.Grey)
                painter.setPen(pen)
            text = f"{amount:,.2f}"
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        painter.setPen(pen)
        painter.restore()

    def createEditor(self, aParent, option, index):
        if index.column() == 1:
            category_selector = CategorySelector(aParent, validate=False)
            return category_selector
        if index.column() == 3:
            tag_selector = TagSelector(aParent, validate=False)
            return tag_selector

    def setModelData(self, editor, model, index):
        if index.column() == 1:
            model.setData(index, editor.selected_id)
            model.setData(index.siblingAtColumn(2), 1) # set confidence level to 1
        if index.column() == 3:
            model.setData(index, editor.selected_id)

#-----------------------------------------------------------------------------------------------------------------------
# Delegate class that shows parameter editor according to its type
class ParameterDelegate(QStyledItemDelegate):    # Code doubles with pieces from delegates.py
    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def createEditor(self, aParent, option, index):
        default_data = index.model().data(index, DEFAULT_DATA_ROLE)
        data_type = type(default_data)
        if data_type == str or data_type == int or data_type == Decimal:
            editor = QLineEdit(aParent)
        elif data_type == QDateTime or data_type == QDate:
            editor = DateTimeEditWithReset(aParent)
            editor.setTimeSpec(Qt.UTC)
            if data_type == QDate:
                editor.setDisplayFormat("dd/MM/yyyy")
            else:
                editor.setDisplayFormat("dd/MM/yyyy hh:mm")
        elif data_type == dict:
            editor = QComboBox(aParent)
            for idx in default_data:
                editor.addItem(default_data[idx], userData=idx)
        else:
            assert False, f"Delegate ParameterDelegate.createEditor() called for unsupported type {data_type}"
        return editor

    def setEditorData(self, editor, index):
        default_data = index.model().data(index, DEFAULT_DATA_ROLE)
        data_type = type(default_data)
        if data_type == str or data_type == int:
            editor.setText(str(index.model().data(index, Qt.EditRole)))
        elif data_type == Decimal:
            editor.setText(localize_decimal(index.model().data(index, Qt.EditRole)))
        elif data_type == QDateTime:
            editor.setDateTime(index.model().data(index, Qt.EditRole))
        elif data_type == QDate:
            editor.setDateTime(QDateTime(index.model().data(index, Qt.EditRole), QTime()))
        elif data_type == dict:
            editor.setCurrentIndex(editor.findData(index.model().data(index, Qt.EditRole)))
        else:
            assert False, f"Delegate ParameterDelegate.setEditorData() called for unsupported type {data_type}"
        return editor

    def setModelData(self, editor, model, index):
        default_data = index.model().data(index, DEFAULT_DATA_ROLE)
        data_type = type(default_data)
        if data_type == str:
            model.setData(index, editor.text())
        elif data_type == int:
            model.setData(index, QLocale().toInt(editor.text())[0])
        elif data_type == QDate:
            model.setData(index, editor.date())
        elif data_type == QDateTime:
            model.setData(index, editor.dateTime())
        elif data_type == Decimal:
            model.setData(index, delocalize_decimal(editor.text()))
        elif data_type == dict:
            model.setData(index, editor.currentData())
        else:
            assert False, f"Delegate ParameterDelegate.setModelData() called for unsupported type {data_type}"

#-----------------------------------------------------------------------------------------------------------------------
# Custom model to display and edit slip lines
class ParamsModel(QAbstractTableModel):
    def __init__(self, params_list: dict, parent=None):
        super().__init__(parent)
        self._params = params_list
        self._values = [(lambda x : x if type(x) != dict else next(iter(x)))(y) for y in self._params.values()]

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        return super().flags(index) | Qt.ItemIsEditable

    def rowCount(self, parent=None):
        return len(self._params)

    def columnCount(self, parent=None):
        return 1

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            value = self._values[index.row()]
            default_value = list(self._params.values())[index.row()]
            if role == Qt.DisplayRole:
                if type(default_value) == Decimal:
                    return str(value)
                elif type(default_value) == dict:
                    return default_value[value]
                else:
                    return value
            elif role == Qt.EditRole:
                return value
            elif role == DEFAULT_DATA_ROLE:
                return default_value
        return None

    def setData(self, index, value, role=Qt.EditRole):
        if index.isValid():
            if role == Qt.EditRole:
                self._values[index.row()] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def headerData(self, row, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Vertical and role == Qt.DisplayRole:
            return list(self._params)[row]
        return None

    def params(self) -> dict:
        return dict(zip(self._params, self._values))


#-----------------------------------------------------------------------------------------------------------------------
class ImportReceiptDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_ImportShopReceiptDlg()
        self.ui.setupUi(self)
        self.model = None
        self.delegate = []
        self.params_model = None
        self._parameter_delegate = ParameterDelegate(self.ui.ReceiptParametersList)
        self.slip_lines = None
        self.receipt_api = None
        self.tensor_flow_present = dependency_present(['tensorflow'])

        self.ui.ScanReceiptQR.clicked.connect(self.processReceiptQR)
        self.ui.DownloadReceiptBtn.clicked.connect(self.processReceiptParams)
        self.ui.AddOperationBtn.clicked.connect(self.addOperation)
        self.ui.ClearBtn.clicked.connect(self.clearSlipData)
        self.ui.AssignCategoryBtn.clicked.connect(self.recognizeCategories)
        self.ui.ReceiptAPICombo.currentIndexChanged.connect(self.change_api)

        self.ui.ReceiptAPICombo.clear()
        for idx, name in ReceiptAPIFactory().supported_names.items():
            self.ui.ReceiptAPICombo.addItem(name, idx)

        self.ui.AssignCategoryBtn.setEnabled(self.tensor_flow_present)

    # -----------------------------------------------------------------------------------------------
    @Slot()
    def change_api(self, index):
        if index < 0:
            return
        api_type = self.ui.ReceiptAPICombo.currentData()
        self.params_model = ParamsModel(ReceiptAPIFactory().get_api_parameters(api_type))
        self.ui.ReceiptParametersList.setModel(self.params_model)
        self.ui.ReceiptParametersList.setItemDelegateForColumn(0, self._parameter_delegate)

    #-----------------------------------------------------------------------------------------------
    # Then it downloads the slip if match found. Otherwise, shows warning message but allows to proceed
    @Slot()
    def processReceiptQR(self):
        scanner = ScanDialog(parent=self, message=self.tr("Please scan main QR code from the receipt"))
        if scanner.exec() != QDialog.Accepted:
            return
        logging.info(self.tr("QR: " + scanner.data))
        try:
            self.receipt_api = ReceiptAPIFactory().get_api_for_qr(scanner.data)
        except ValueError as e:
            logging.warning(e)
            return
        self.receipt_api.slip_load_ok.connect(self.slip_loaded)
        self.downloadSlipJSON()

    @Slot()
    def processReceiptParams(self):
        api_type = self.ui.ReceiptAPICombo.currentData()
        try:
            self.receipt_api = ReceiptAPIFactory().get_api_with_params(api_type, self.params_model.params())
        except ValueError as e:
            logging.warning(e)
            return
        self.receipt_api.slip_load_ok.connect(self.slip_loaded)
        self.downloadSlipJSON()

    def downloadSlipJSON(self):
        if self.receipt_api is None:
            return
        if not self.receipt_api.activate_session():
            return
        self.receipt_api.query_slip()

    def slip_loaded(self):
        self.parseJSON()

    def parseJSON(self):
        self.slip_lines = pd.DataFrame(self.receipt_api.slip_lines())
        self.ui.SlipShopName.setText(self.receipt_api.shop_name())
        peer_id = JalPeer.get_id_by_mapped_name(self.ui.SlipShopName.text())
        if peer_id is not None:
            self.ui.PeerEdit.selected_id = peer_id
        self.ui.SlipDateTime.setDateTime(self.receipt_api.datetime())
        # Assign empty category
        self.slip_lines['category'] = 0
        self.slip_lines['confidence'] = 1
        # Assign empty tags
        self.slip_lines['tag'] = None
        self.slip_lines = self.slip_lines[['name', 'category', 'confidence', 'tag', 'amount']]

        self.model = PandasLinesModel(self.slip_lines, self)
        self.ui.LinesTableView.setModel(self.model)

        self.delegate = SlipLinesDelegate(self.ui.LinesTableView)
        for column in range(self.model.columnCount()):
            if column == 0:
                self.ui.LinesTableView.horizontalHeader().setSectionResizeMode(column, QHeaderView.Stretch)
            elif column == 1:
                self.ui.LinesTableView.setColumnWidth(column, 200)
            elif column == 2:
                self.ui.LinesTableView.setColumnHidden(column, True)
            else:
                self.ui.LinesTableView.setColumnWidth(column, 100)
            self.ui.LinesTableView.setItemDelegateForColumn(column, self.delegate)
        font = self.ui.LinesTableView.horizontalHeader().font()
        font.setBold(True)
        self.ui.LinesTableView.horizontalHeader().setFont(font)
        self.ui.LinesTableView.show()
        self.recognizeCategories()

    def addOperation(self):
        if self.ui.AccountEdit.selected_id == 0:
            logging.warning(self.tr("Not possible to import receipt: no account set for import"))
            return
        if self.ui.PeerEdit.selected_id == 0:
            logging.warning(self.tr("Not possible to import receipt: can't import: no peer set for import"))
            return
        if self.slip_lines[self.slip_lines['category'] == 0].shape[0] != 0:
            logging.warning(self.tr("Not possible to import receipt: some categories are not set"))
            return

        details = []
        for index, row in self.slip_lines.iterrows():
            details.append({
                "category_id": row['category'],
                "tag_id": row['tag'],
                "amount": row['amount'],
                "note": row['name']
            })
            JalCategory(row['category']).add_or_update_mapped_name(row['name'])
        operation = {
            "timestamp": self.ui.SlipDateTime.dateTime().toSecsSinceEpoch(),
            "account_id": self.ui.AccountEdit.selected_id,
            "peer_id": self.ui.PeerEdit.selected_id,
            "lines": details
        }
        LedgerTransaction.create_new(LedgerTransaction.IncomeSpending, operation)
        JalPeer(self.ui.PeerEdit.selected_id).add_or_update_mapped_name(self.ui.SlipShopName.text(), )
        self.clearSlipData()

    def clearSlipData(self):
        self.slip_lines = None
        self.ui.LinesTableView.setModel(None)

    @Slot()
    def recognizeCategories(self):
        if not self.tensor_flow_present:
            logging.warning(self.tr("Categories are not recognized: Tensorflow is not found"))
            return
        self.slip_lines['category'], self.slip_lines['confidence'] = \
            recognize_categories(self.slip_lines['name'].tolist())
        self.model.dataChanged.emit(None, None)  # refresh full view



from datetime import datetime
from dateutil import tz

from jal.ui.widgets.ui_trade_operation import Ui_TradeOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.delegates import WidgetMapperDelegateBase
from jal.db.operations import LedgerTransaction


# ----------------------------------------------------------------------------------------------------------------------
class TradeWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate,
                          'settlement': self.timestamp_delegate,
                          'asset_id': self.symbol_delegate,
                          'qty': self.decimal_long_delegate,
                          'price': self.decimal_long_delegate,
                          'fee': self.decimal_long_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class TradeWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_TradeOperation)
        self.operation_type = LedgerTransaction.Trade
        super()._init_db("trades")
        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.settlement_editor.setFixedWidth(self.ui.settlement_editor.fontMetrics().horizontalAdvance("00/00/0000") * 1.5)

        self.mapper.setItemDelegate(TradeWidgetDelegate(self.mapper))

        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.asset_widget.changed.connect(self.mapper.submit)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.settlement_editor, self.model.fieldIndex("settlement"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.currency_price, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.currency_fee, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.asset_widget, self.model.fieldIndex("asset_id"))
        self.mapper.addMapping(self.ui.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.ui.qty_edit, self.model.fieldIndex("qty"))
        self.mapper.addMapping(self.ui.price_edit, self.model.fieldIndex("price"))
        self.mapper.addMapping(self.ui.fee_edit, self.model.fieldIndex("fee"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.model.select()

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("settlement", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        new_record.setValue("account_id", account_id)
        new_record.setValue("asset_id", 0)
        new_record.setValue("qty", '0')
        new_record.setValue("price", '0')
        new_record.setValue("fee", '0')
        new_record.setValue("note", None)
        return new_record

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("settlement", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        return new_record


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'login_pingo_doce_dlg.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QFormLayout, QFrame,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QVBoxLayout, QWidget)

class Ui_LoginPingoDoceDialog(object):
    def setupUi(self, LoginPingoDoceDialog):
        if not LoginPingoDoceDialog.objectName():
            LoginPingoDoceDialog.setObjectName(u"LoginPingoDoceDialog")
        LoginPingoDoceDialog.resize(296, 130)
        self.verticalLayout_3 = QVBoxLayout(LoginPingoDoceDialog)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(2, 2, 2, 2)
        self.PhoneNumberFrame = QFrame(LoginPingoDoceDialog)
        self.PhoneNumberFrame.setObjectName(u"PhoneNumberFrame")
        self.PhoneNumberFrame.setFrameShape(QFrame.NoFrame)
        self.PhoneNumberFrame.setFrameShadow(QFrame.Plain)
        self.formLayout_2 = QFormLayout(self.PhoneNumberFrame)
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.formLayout_2.setContentsMargins(6, -1, 6, 0)
        self.PhoneLbl = QLabel(self.PhoneNumberFrame)
        self.PhoneLbl.setObjectName(u"PhoneLbl")

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.PhoneLbl)

        self.PhoneNumberEdit = QLineEdit(self.PhoneNumberFrame)
        self.PhoneNumberEdit.setObjectName(u"PhoneNumberEdit")
        self.PhoneNumberEdit.setInputMask(u"+351-999-999-999;_")

        self.formLayout_2.setWidget(0, QFormLayout.FieldRole, self.PhoneNumberEdit)


        self.verticalLayout_3.addWidget(self.PhoneNumberFrame)

        self.PasswordFrame = QFrame(LoginPingoDoceDialog)
        self.PasswordFrame.setObjectName(u"PasswordFrame")
        self.PasswordFrame.setFrameShape(QFrame.NoFrame)
        self.PasswordFrame.setFrameShadow(QFrame.Plain)
        self.formLayout_3 = QFormLayout(self.PasswordFrame)
        self.formLayout_3.setObjectName(u"formLayout_3")
        self.formLayout_3.setContentsMargins(6, -1, 6, 0)
        self.PasswordLbl = QLabel(self.PasswordFrame)
        self.PasswordLbl.setObjectName(u"PasswordLbl")

        self.formLayout_3.setWidget(0, QFormLayout.LabelRole, self.PasswordLbl)

        self.PasswordEdit = QLineEdit(self.PasswordFrame)
        self.PasswordEdit.setObjectName(u"PasswordEdit")
        self.PasswordEdit.setEchoMode(QLineEdit.Password)

        self.formLayout_3.setWidget(0, QFormLayout.FieldRole, self.PasswordEdit)


        self.verticalLayout_3.addWidget(self.PasswordFrame)

        self.frame = QFrame(LoginPingoDoceDialog)
        self.frame.setObjectName(u"frame")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Plain)
        self.verticalLayout_9 = QVBoxLayout(self.frame)
        self.verticalLayout_9.setObjectName(u"verticalLayout_9")
        self.LoginBtn = QPushButton(self.frame)
        self.LoginBtn.setObjectName(u"LoginBtn")
        sizePolicy1 = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.LoginBtn.sizePolicy().hasHeightForWidth())
        self.LoginBtn.setSizePolicy(sizePolicy1)

        self.verticalLayout_9.addWidget(self.LoginBtn, 0, Qt.AlignHCenter|Qt.AlignTop)


        self.verticalLayout_3.addWidget(self.frame)


        self.retranslateUi(LoginPingoDoceDialog)

        QMetaObject.connectSlotsByName(LoginPingoDoceDialog)
    # setupUi

    def retranslateUi(self, LoginPingoDoceDialog):
        LoginPingoDoceDialog.setWindowTitle(QCoreApplication.translate("LoginPingoDoceDialog", u"Authorization Pingo Doce", None))
        self.PhoneLbl.setText(QCoreApplication.translate("LoginPingoDoceDialog", u"Phone number:", None))
        self.PhoneNumberEdit.setText(QCoreApplication.translate("LoginPingoDoceDialog", u"+351---", None))
        self.PasswordLbl.setText(QCoreApplication.translate("LoginPingoDoceDialog", u"Password:", None))
        self.PasswordEdit.setInputMask("")
        self.PasswordEdit.setText("")
        self.LoginBtn.setText(QCoreApplication.translate("LoginPingoDoceDialog", u"Login", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'corporate_action_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDateTimeEdit, QGridLayout,
    QHeaderView, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpacerItem, QWidget)

from jal.widgets.custom.tableview_with_footer import TableViewWithFooter
from jal.widgets.reference_selector import (AccountSelector, AssetSelector)

class Ui_CorporateActionOperation(object):
    def setupUi(self, CorporateActionOperation):
        if not CorporateActionOperation.objectName():
            CorporateActionOperation.setObjectName(u"CorporateActionOperation")
        CorporateActionOperation.resize(969, 244)
        self.layout = QGridLayout(CorporateActionOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.add_button = QPushButton(CorporateActionOperation)
        self.add_button.setObjectName(u"add_button")

        self.layout.addWidget(self.add_button, 2, 7, 1, 1)

        self.asset_label = QLabel(CorporateActionOperation)
        self.asset_label.setObjectName(u"asset_label")

        self.layout.addWidget(self.asset_label, 3, 3, 1, 1)

        self.qty_edit = QLineEdit(CorporateActionOperation)
        self.qty_edit.setObjectName(u"qty_edit")

        self.layout.addWidget(self.qty_edit, 4, 4, 1, 1)

        self.account_label = QLabel(CorporateActionOperation)
        self.account_label.setObjectName(u"account_label")

        self.layout.addWidget(self.account_label, 2, 3, 1, 1)

        self.note_label = QLabel(CorporateActionOperation)
        self.note_label.setObjectName(u"note_label")

        self.layout.addWidget(self.note_label, 5, 0, 1, 1)

        self.commit_button = QPushButton(CorporateActionOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 7, 1, 1)

        self.qty_label = QLabel(CorporateActionOperation)
        self.qty_label.setObjectName(u"qty_label")

        self.layout.addWidget(self.qty_label, 4, 3, 1, 1)

        self.main_label = QLabel(CorporateActionOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 3)

        self.date_label = QLabel(CorporateActionOperation)
        self.date_label.setObjectName(u"date_label")

        self.layout.addWidget(self.date_label, 2, 0, 1, 1)

        self.timestamp_editor = QDateTimeEdit(CorporateActionOperation)
        self.timestamp_editor.setObjectName(u"timestamp_editor")
        self.timestamp_editor.setCalendarPopup(True)
        self.timestamp_editor.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.timestamp_editor, 2, 2, 1, 1)

        self.arrow = QLabel(CorporateActionOperation)
        self.arrow.setObjectName(u"arrow")

        self.layout.addWidget(self.arrow, 3, 5, 2, 1)

        self.type_label = QLabel(CorporateActionOperation)
        self.type_label.setObjectName(u"type_label")

        self.layout.addWidget(self.type_label, 3, 0, 1, 1)

        self.asset_widget = AssetSelector(CorporateActionOperation)
        self.asset_widget.setObjectName(u"asset_widget")

        self.layout.addWidget(self.asset_widget, 3, 4, 1, 1)

        self.account_widget = AccountSelector(CorporateActionOperation)
        self.account_widget.setObjectName(u"account_widget")

        self.layout.addWidget(self.account_widget, 2, 4, 1, 1)

        self.number = QLineEdit(CorporateActionOperation)
        self.number.setObjectName(u"number")

        self.layout.addWidget(self.number, 4, 2, 1, 1)

        self.type = QComboBox(CorporateActionOperation)
        self.type.setObjectName(u"type")

        self.layout.addWidget(self.type, 3, 2, 1, 1)

        self.revert_button = QPushButton(CorporateActionOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 8, 1, 1)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.vertical_spacer, 6, 0, 1, 1)

        self.note = QLineEdit(CorporateActionOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 5, 2, 1, 3)

        self.del_button = QPushButton(CorporateActionOperation)
        self.del_button.setObjectName(u"del_button")

        self.layout.addWidget(self.del_button, 2, 8, 1, 1)

        self.number_label = QLabel(CorporateActionOperation)
        self.number_label.setObjectName(u"number_label")

        self.layout.addWidget(self.number_label, 4, 0, 1, 1)

        self.results_table = TableViewWithFooter(CorporateActionOperation)
        self.results_table.setObjectName(u"results_table")
        self.results_table.setAlternatingRowColors(True)
        self.results_table.verticalHeader().setVisible(False)
        self.results_table.verticalHeader().setMinimumSectionSize(20)
        self.results_table.verticalHeader().setDefaultSectionSize(20)

        self.layout.addWidget(self.results_table, 3, 6, 3, 3)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontalSpacer, 2, 6, 1, 1)


        self.retranslateUi(CorporateActionOperation)

        QMetaObject.connectSlotsByName(CorporateActionOperation)
    # setupUi

    def retranslateUi(self, CorporateActionOperation):
        CorporateActionOperation.setWindowTitle(QCoreApplication.translate("CorporateActionOperation", u"Form", None))
#if QT_CONFIG(tooltip)
        self.add_button.setToolTip(QCoreApplication.translate("CorporateActionOperation", u"Add asset", None))
#endif // QT_CONFIG(tooltip)
        self.add_button.setText("")
        self.asset_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Asset", None))
        self.account_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Account", None))
        self.note_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Note", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("CorporateActionOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
        self.qty_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Qty", None))
        self.main_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Corporate Action", None))
        self.date_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Date/Time", None))
        self.timestamp_editor.setDisplayFormat(QCoreApplication.translate("CorporateActionOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.arrow.setText(QCoreApplication.translate("CorporateActionOperation", u" - ", None))
        self.type_label.setText(QCoreApplication.translate("CorporateActionOperation", u"Type", None))
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("CorporateActionOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
#if QT_CONFIG(tooltip)
        self.del_button.setToolTip(QCoreApplication.translate("CorporateActionOperation", u"Remove asset", None))
#endif // QT_CONFIG(tooltip)
        self.del_button.setText("")
        self.number_label.setText(QCoreApplication.translate("CorporateActionOperation", u"#", None))
    # retranslateUi




from datetime import datetime
from dateutil import tz

from PySide6.QtCore import Slot, QStringListModel, QByteArray
from PySide6.QtWidgets import QMessageBox
from jal.ui.widgets.ui_dividend_operation import Ui_DividendOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.delegates import WidgetMapperDelegateBase
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.helpers import db_row2dict
from jal.db.operations import LedgerTransaction, Dividend


# ----------------------------------------------------------------------------------------------------------------------
class DividendWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate,
                          'ex_date': self.timestamp_delegate,
                          'asset_id': self.symbol_delegate,
                          'amount': self.decimal_delegate,
                          'tax': self.decimal_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class DividendWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_DividendOperation)
        self.operation_type = LedgerTransaction.Dividend
        super()._init_db("dividends")
        self.combo_model = QStringListModel([self.tr("N/A"),
                                             self.tr("Dividend"),
                                             self.tr("Bond Interest"),
                                             self.tr("Stock Dividend"),
                                             self.tr("Stock Vesting")])
        self.ui.type.setModel(self.combo_model)
        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.ex_date_editor.setFixedWidth(self.ui.ex_date_editor.fontMetrics().horizontalAdvance("00/00/0000") * 1.5)
        self.ui.price_label.setVisible(False)
        self.ui.price_edit.setVisible(False)

        self.mapper.setItemDelegate(DividendWidgetDelegate(self.mapper))

        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.asset_widget.changed.connect(self.assetChanged)
        self.ui.type.currentIndexChanged.connect(self.typeChanged)
        self.ui.timestamp_editor.dateTimeChanged.connect(self.refreshAssetPrice)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.ex_date_editor, self.model.fieldIndex("ex_date"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.currency, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.asset_widget, self.model.fieldIndex("asset_id"))
        self.mapper.addMapping(self.ui.type, self.model.fieldIndex("type"), QByteArray().setRawData("currentIndex", 12))
        self.mapper.addMapping(self.ui.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.ui.dividend_edit, self.model.fieldIndex("amount"))
        self.mapper.addMapping(self.ui.tax_edit, self.model.fieldIndex("tax"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.model.select()

    @Slot()
    def assetChanged(self):
        self.mapper.submit()
        self.refreshAssetPrice()

    @Slot()
    def typeChanged(self, dividend_type_id):
        self.ui.price_label.setVisible(
            dividend_type_id == Dividend.StockDividend or dividend_type_id == Dividend.StockVesting)
        self.ui.price_edit.setVisible(
            dividend_type_id == Dividend.StockDividend or dividend_type_id == Dividend.StockVesting)
        self.refreshAssetPrice()

    def refreshAssetPrice(self):
        if self.ui.type.currentIndex() == Dividend.StockDividend or self.ui.type.currentIndex() == Dividend.StockVesting:
            dividend_timestamp = self.ui.timestamp_editor.dateTime().toSecsSinceEpoch()
            timestamp, price = JalAsset(self.ui.asset_widget.selected_id).quote(dividend_timestamp,
                                                                             JalAccount(self.ui.account_widget.selected_id).currency())
            if timestamp == dividend_timestamp:
                self.ui.price_edit.setText(str(price))
                self.ui.price_edit.setStyleSheet('')
                self.ui.price_edit.setToolTip("")
            else:
                self.ui.price_edit.setText(self.tr("No quote"))
                self.ui.price_edit.setStyleSheet("color: red")
                self.ui.price_edit.setToolTip(
                    self.tr("You should set quote via Data->Quotes menu for Date/Time of the dividend"))

    def _validated(self):
        fields = db_row2dict(self.model, 0)
        if not fields['type']:
            QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Please set a type of the dividend."), QMessageBox.Ok)
            return False
        return True

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("ex_date", 0)
        new_record.setValue("type", 0)
        new_record.setValue("number", '')
        new_record.setValue("account_id", account_id)
        new_record.setValue("asset_id", 0)
        new_record.setValue("amount", '0')
        new_record.setValue("tax", '0')
        new_record.setValue("note", None)
        return new_record

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("ex_date", 0)
        new_record.setValue("number", '')
        return new_record
